"""
Reduces scattering data into 2D and 1D datasets.
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D
from matplotlib.markers import MarkerStyle
from matplotlib.ticker import MultipleLocator
from matplotlib import colors
from matplotlib import patches
from IPython.display import display, Markdown
from nexusformat.nexus import NXfield, NXdata, nxload, NeXusError, NXroot, NXentry, nxsave
from scipy import ndimage

# Specify items on which users are allowed to perform standalone imports
__all__ = ['load_data', 'load_transform', 'plot_slice', 'Scissors',
           'reciprocal_lattice_params', 'rotate_data',
           'array_to_nxdata', 'Padder']


def load_data(path):
    """
    Load data from a NeXus file at a specified path. It is assumed that the data follows the CHESS
    file structure (i.e., root/entry/data/counts, etc.).

    Parameters
    ----------
    path : str
        The path to the NeXus data file.

    Returns
    -------
    data : nxdata object
        The loaded data stored in a nxdata object.

    """

    g = nxload(path)
    try:
        print(g.entry.data.tree)
    except NeXusError:
        pass

    return g.entry.data


def load_transform(path):
    """
    Load data obtained from nxrefine output from a specified path.

    Parameters
    ----------
    path : str The path to the transform data file.

    Returns
    -------
    data : nxdata object The loaded data stored in a nxdata object.
    """
    g = nxload(path)
    return NXdata(NXfield(g.entry.transform.data.nxdata.transpose(2, 1, 0), name='counts'),
                  (g.entry.transform.Qh, g.entry.transform.Qk, g.entry.transform.Ql))


def array_to_nxdata(array, data_template, signal_name='counts'):
    """
    Create an NXdata object from an input array and an NXdata template,
    with an optional signal name.

    Parameters
    ----------
    array : array-like
        The data array to be included in the NXdata object.

    data_template : NXdata
        An NXdata object serving as a template, which provides information
        about axes and other metadata.

    signal_name : str, optional
        The name of the signal within the NXdata object. If not provided,
        the default signal name 'counts' is used.

    Returns
    -------
    NXdata
        An NXdata object containing the input data array and associated axes
        based on the template.
    """
    d = data_template
    return NXdata(NXfield(array, name=signal_name),
                  tuple(d[d.axes[i]] for i in range(len(d.axes))))


def plot_slice(data, X=None, Y=None, transpose=False, vmin=None, vmax=None,
               skew_angle=90, ax=None, xlim=None, ylim=None,
               xticks=None, yticks=None, cbar=True, logscale=False,
               symlogscale=False, cmap='viridis', linthresh=1,
               title=None, mdheading=None, cbartitle=None,
               **kwargs):
    """
    Plot a 2D slice of the provided dataset, with optional transformations
    and customizations.

    Parameters
    ----------
    data : :class:`nexusformat.nexus.NXdata` or ndarray
        The dataset to plot. Can be an `NXdata` object or a `numpy` array.

    X : NXfield, optional
        The X axis values. If None, a default range from 0 to the number of
         columns in `data` is used.

    Y : NXfield, optional
        The Y axis values. If None, a default range from 0 to the number of
         rows in `data` is used.

    transpose : bool, optional
        If True, transpose the dataset and its axes before plotting.
        Default is False.

    vmin : float, optional
        The minimum value for the color scale. If not provided, the minimum
         value of the dataset is used.

    vmax : float, optional
        The maximum value for the color scale. If not provided, the maximum
         value of the dataset is used.

    skew_angle : float, optional
        The angle in degrees to shear the plot. Default is 90 degrees (no skew).

    ax : matplotlib.axes.Axes, optional
        The `matplotlib` axis to plot on. If None, a new figure and axis will
         be created.

    xlim : tuple, optional
        The limits for the x-axis. If None, the limits are set automatically
         based on the data.

    ylim : tuple, optional
        The limits for the y-axis. If None, the limits are set automatically
         based on the data.

    xticks : float or list of float, optional
        The major tick interval or specific tick locations for the x-axis.
         Default is to use a minor tick interval of 1.

    yticks : float or list of float, optional
        The major tick interval or specific tick locations for the y-axis.
        Default is to use a minor tick interval of 1.

    cbar : bool, optional
        Whether to include a colorbar. Default is True.

    logscale : bool, optional
        Whether to use a logarithmic color scale. Default is False.

    symlogscale : bool, optional
        Whether to use a symmetrical logarithmic color scale. Default is False.

    cmap : str or Colormap, optional
        The colormap to use for the plot. Default is 'viridis'.

    linthresh : float, optional
        The linear threshold for symmetrical logarithmic scaling. Default is 1.

    title : str, optional
        The title for the plot. If None, no title is set.

    mdheading : str, optional
        A Markdown heading to display above the plot. If 'None' or not provided,
         no heading is displayed.

    cbartitle : str, optional
        The title for the colorbar. If None, the colorbar label will be set to
         the name of the signal.

    **kwargs
        Additional keyword arguments passed to `pcolormesh`.

    Returns
    -------
    p : :class:`matplotlib.collections.QuadMesh`
        The `matplotlib` QuadMesh object representing the plotted data.
    """
    if isinstance(data, np.ndarray):
        if X is None:
            X = NXfield(np.linspace(0, data.shape[1], data.shape[1]), name='x')
        if Y is None:
            Y = NXfield(np.linspace(0, data.shape[0], data.shape[0]), name='y')
        if transpose:
            X, Y = Y, X
            data = data.transpose()
        data = NXdata(NXfield(data, name='value'), (X, Y))
        data_arr = data
    elif isinstance(data, (NXdata, NXfield)):
        if X is None:
            X = data[data.axes[0]]
        if Y is None:
            Y = data[data.axes[1]]
        if transpose:
            X, Y = Y, X
            data = data.transpose()
        data_arr = data[data.signal].nxdata.transpose()
    else:
        raise TypeError(f"Unexpected data type: {type(data)}. "
                        f"Supported types are np.ndarray and NXdata.")

    # Display Markdown heading
    if mdheading is None:
        pass
    elif mdheading == "None":
        display(Markdown('### Figure'))
    else:
        display(Markdown('### Figure - ' + mdheading))

    # Inherit axes if user provides some
    if ax is not None:
        fig = ax.get_figure()
    # Otherwise set up some default axes
    else:
        fig = plt.figure()
        ax = fig.add_axes([0, 0, 1, 1])

    # If limits not provided, use extrema
    if vmin is None:
        vmin = data_arr.min()
    if vmax is None:
        vmax = data_arr.max()

    # Set norm (linear scale, logscale, or symlogscale)
    norm = colors.Normalize(vmin=vmin, vmax=vmax)  # Default: linear scale

    if symlogscale:
        norm = colors.SymLogNorm(linthresh=linthresh, vmin=-1 * vmax, vmax=vmax)
    elif logscale:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)

    # Plot data
    p = ax.pcolormesh(X.nxdata, Y.nxdata, data_arr, shading='auto', norm=norm, cmap=cmap, **kwargs)

    ## Transform data to new coordinate system if necessary
    # Correct skew angle
    skew_angle_adj = 90 - skew_angle
    # Create blank 2D affine transformation
    t = Affine2D()
    # Scale y-axis to preserve norm while shearing
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180))
    # Shear along x-axis
    t += Affine2D().skew_deg(skew_angle_adj, 0)
    # Return to original y-axis scaling
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180)).inverted()
    ## Correct for x-displacement after shearing
    # If ylims provided, use those
    if ylim is not None:
        # Set ylims
        ax.set(ylim=ylim)
        ymin, ymax = ylim
    # Else, use current ylims
    else:
        ymin, ymax = ax.get_ylim()
    # Use ylims to calculate translation (necessary to display axes in correct position)
    p.set_transform(t
                    + Affine2D().translate(-ymin * np.sin(skew_angle_adj * np.pi / 180), 0)
                    + ax.transData)

    # Set x limits
    if xlim is not None:
        xmin, xmax = xlim
    else:
        xmin, xmax = ax.get_xlim()
    if skew_angle <= 90:
        ax.set(xlim=(xmin, xmax + (ymax - ymin) / np.tan((90 - skew_angle_adj) * np.pi / 180)))
    else:
        ax.set(xlim=(xmin - (ymax - ymin) / np.tan((skew_angle_adj - 90) * np.pi / 180), xmax))

    # Correct aspect ratio for the x/y axes after transformation
    ax.set(aspect=np.cos(skew_angle_adj * np.pi / 180))

    # Add tick marks all around
    ax.tick_params(direction='in', top=True, right=True, which='both')

    # Set tick locations
    if xticks is None:
        # Add default minor ticks
        ax.xaxis.set_minor_locator(MultipleLocator(1))
    else:
        # Otherwise use user provided values
        ax.xaxis.set_major_locator(MultipleLocator(xticks))
        ax.xaxis.set_minor_locator(MultipleLocator(1))
    if yticks is None:
        # Add default minor ticks
        ax.yaxis.set_minor_locator(MultipleLocator(1))
    else:
        # Otherwise use user provided values
        ax.yaxis.set_major_locator(MultipleLocator(yticks))
        ax.yaxis.set_minor_locator(MultipleLocator(1))

    # Apply transform to tick marks
    for i in range(0, len(ax.xaxis.get_ticklines())):
        # Tick marker
        m = MarkerStyle(3)
        line = ax.xaxis.get_majorticklines()[i]
        if i % 2:
            # Top ticks (translation here makes their direction="in")
            m._transform.set(Affine2D().translate(0, -1) + Affine2D().skew_deg(skew_angle_adj, 0))
            # This first method shifts the top ticks horizontally to match the skew angle.
            # This does not look good in all cases.
            # line.set_transform(Affine2D().translate((ymax-ymin)*np.sin(skew_angle*np.pi/180),0) +
            #     line.get_transform())
            # This second method skews the tick marks in place and
            # can sometimes lead to them being misaligned.
            line.set_transform(line.get_transform())  # This does nothing
        else:
            # Bottom ticks
            m._transform.set(Affine2D().skew_deg(skew_angle_adj, 0))

        line.set_marker(m)

    for i in range(0, len(ax.xaxis.get_minorticklines())):
        m = MarkerStyle(2)
        line = ax.xaxis.get_minorticklines()[i]
        if i % 2:
            m._transform.set(Affine2D().translate(0, -1) + Affine2D().skew_deg(skew_angle_adj, 0))
        else:
            m._transform.set(Affine2D().skew_deg(skew_angle_adj, 0))

        line.set_marker(m)

    if cbar:
        colorbar = fig.colorbar(p)
        if cbartitle is None:
            colorbar.set_label(data.signal)

    ax.set(
        xlabel=X.nxname,
        ylabel=Y.nxname,
    )

    if title is not None:
        ax.set_title(title)

    # Return the quadmesh object
    return p


class Scissors:
    """
    Scissors class provides functionality for reducing data to a 1D linecut using an integration
    window.

    Attributes
    ----------
    data : :class:`nexusformat.nexus.NXdata` or None
        Input :class:`nexusformat.nexus.NXdata`.
    center : tuple or None
        Central coordinate around which to perform the linecut.
    window : tuple or None
        Extents of the window for integration along each axis.
    axis : int or None
        Axis along which to perform the integration.
    integration_volume : :class:`nexusformat.nexus.NXdata` or None
        Data array after applying the integration window.
    integrated_axes : tuple or None
        Indices of axes that were integrated.
    linecut : :class:`nexusformat.nexus.NXdata` or None
        1D linecut data after integration.
    integration_window : tuple or None
        Slice object representing the integration window in the data array.

    Methods
    -------
    set_data(data)
        Set the input :class:`nexusformat.nexus.NXdata`.
    get_data()
        Get the input :class:`nexusformat.nexus.NXdata`.
    set_center(center)
        Set the central coordinate for the linecut.
    set_window(window, axis=None, verbose=False)
        Set the extents of the integration window.
    get_window()
        Get the extents of the integration window.
    cut_data(center=None, window=None, axis=None, verbose=False)
        Reduce data to a 1D linecut using the integration window.
    highlight_integration_window(data=None, label=None, highlight_color='red', **kwargs)
        Plot the integration window highlighted on a 2D heatmap of the full dataset.
    plot_integration_window(**kwargs)
        Plot a 2D heatmap of the integration window data.
    """

    def __init__(self, data=None, center=None, window=None, axis=None):
        """
        Initializes a Scissors object.

        Parameters
        ----------
        data : :class:`nexusformat.nexus.NXdata` or None, optional
            Input NXdata. Default is None.
        center : tuple or None, optional
            Central coordinate around which to perform the linecut. Default is None.
        window : tuple or None, optional
            Extents of the window for integration along each axis. Default is None.
        axis : int or None, optional
            Axis along which to perform the integration. Default is None.
        """

        self.data = data
        self.center = tuple(float(i) for i in center) if center is not None else None
        self.window = tuple(float(i) for i in window) if window is not None else None
        self.axis = axis

        self.integration_volume = None
        self.integrated_axes = None
        self.linecut = None
        self.integration_window = None

    def set_data(self, data):
        """
        Set the input NXdata.

        Parameters
        ----------
        data : :class:`nexusformat.nexus.NXdata`
            Input data array.
        """
        self.data = data

    def get_data(self):
        """
        Get the input data array.

        Returns
        -------
        ndarray or None
            Input data array.
        """
        return self.data

    def set_center(self, center):
        """
        Set the central coordinate for the linecut.

        Parameters
        ----------
        center : tuple
            Central coordinate around which to perform the linecut.
        """
        self.center = tuple(float(i) for i in center) if center is not None else None

    def set_window(self, window, axis=None, verbose=False):
        """
        Set the extents of the integration window.

        Parameters
        ----------
        window : tuple
            Extents of the window for integration along each axis.
        axis : int or None, optional
            The axis along which to perform the linecut. If not specified, the value from the
            object's attribute will be used.
        verbose : bool, optional
            Enables printout of linecut axis and integrated axes. Default False.

        """
        self.window = tuple(float(i) for i in window) if window is not None else None

        # Determine the axis for integration
        self.axis = window.index(max(window)) if axis is None else axis

        # Determine the integrated axes (axes other than the integration axis)
        self.integrated_axes = tuple(i for i in range(self.data.ndim) if i != self.axis)

        if verbose:
            print("Linecut axis: " + str(self.data.axes[self.axis]))
            print("Integrated axes: " + str([self.data.axes[axis]
                                             for axis in self.integrated_axes]))

    def get_window(self):
        """
        Get the extents of the integration window.

        Returns
        -------
        tuple or None
            Extents of the integration window.
        """
        return self.window

    def cut_data(self, center=None, window=None, axis=None, verbose=False):
        """
        Reduces data to a 1D linecut with integration extents specified by the
        window about a central coordinate.

        Parameters
        ----------
        center : float or None, optional
            Central coordinate for the linecut. If not specified, the value from the object's
            attribute will be used.
        window : tuple or None, optional
            Integration window extents around the central coordinate. If not specified, the value
            from the object's attribute will be used.
        axis : int or None, optional
            The axis along which to perform the linecut. If not specified, the value from the
            object's attribute will be used.
        verbose : bool
            Enables printout of linecut axis and integrated axes. Default False.

        Returns
        -------
        integrated_data : :class:`nexusformat.nexus.NXdata`
            1D linecut data after integration.

        """

        # Extract necessary attributes from the object
        data = self.data
        center = center if center is not None else self.center
        self.set_center(center)
        window = window if window is not None else self.window
        self.set_window(window, axis, verbose)

        # Convert the center to a tuple of floats
        center = tuple(float(c) for c in center)

        # Calculate the start and stop indices for slicing the data
        start = np.subtract(center, window)
        stop = np.add(center, window)
        slice_obj = tuple(slice(s, e) for s, e in zip(start, stop))
        self.integration_window = slice_obj

        # Perform the data cut
        self.integration_volume = data[slice_obj]
        self.integration_volume.nxname = data.nxname

        # Perform integration along the integrated axes
        integrated_data = np.sum(self.integration_volume[self.integration_volume.signal].nxdata,
                                 axis=self.integrated_axes)

        # Create an NXdata object for the linecut data
        self.linecut = NXdata(NXfield(integrated_data, name=self.integration_volume.signal),
                              self.integration_volume[self.integration_volume.axes[self.axis]])
        self.linecut.nxname = self.integration_volume.nxname

        return self.linecut

    def highlight_integration_window(self, data=None, label=None, highlight_color='red', **kwargs):
        """
        Plots integration window highlighted on the three principal cross sections of the first
        temperature dataset.

        Parameters
        ----------
        data : array-like, optional
            The 2D heatmap dataset to plot. If not provided, the dataset stored in `self.data` will
            be used.
        label : str, optional
            The label for the integration window plot.
        highlight_color : str, optional
            The edge color used to highlight the integration window. Default is 'red'.
        **kwargs : keyword arguments, optional
            Additional keyword arguments to customize the plot.

        """
        data = self.data if data is None else data
        center = self.center
        window = self.window

        # Create a figure and subplots
        fig, axes = plt.subplots(1, 3, figsize=(15, 4))

        # Plot cross section 1
        slice_obj = [slice(None)] * data.ndim
        slice_obj[2] = center[2]

        p1 = plot_slice(data[slice_obj],
                        X=data[data.axes[0]],
                        Y=data[data.axes[1]],
                        ax=axes[0],
                        **kwargs)
        ax = axes[0]
        rect_diffuse = patches.Rectangle(
            (center[0] - window[0],
             center[1] - window[1]),
            2 * window[0], 2 * window[1],
            linewidth=1, edgecolor=highlight_color,
            facecolor='none', transform=p1.get_transform(), label=label,
        )
        ax.add_patch(rect_diffuse)

        # Plot cross section 2
        slice_obj = [slice(None)] * data.ndim
        slice_obj[1] = center[1]

        p2 = plot_slice(data[slice_obj],
                        X=data[data.axes[0]],
                        Y=data[data.axes[2]],
                        ax=axes[1],
                        **kwargs)
        ax = axes[1]
        rect_diffuse = patches.Rectangle(
            (center[0] - window[0],
             center[2] - window[2]),
            2 * window[0], 2 * window[2],
            linewidth=1, edgecolor=highlight_color,
            facecolor='none', transform=p2.get_transform(), label=label,
        )
        ax.add_patch(rect_diffuse)

        # Plot cross section 3
        slice_obj = [slice(None)] * data.ndim
        slice_obj[0] = center[0]

        p3 = plot_slice(data[slice_obj],
                        X=data[data.axes[1]],
                        Y=data[data.axes[2]],
                        ax=axes[2],
                        **kwargs)
        ax = axes[2]
        rect_diffuse = patches.Rectangle(
            (center[1] - window[1],
             center[2] - window[2]),
            2 * window[1], 2 * window[2],
            linewidth=1, edgecolor=highlight_color,
            facecolor='none', transform=p3.get_transform(), label=label,
        )
        ax.add_patch(rect_diffuse)

        # Adjust subplot padding
        fig.subplots_adjust(wspace=0.5)

        if label is not None:
            [ax.legend() for ax in axes]

        plt.show()

        return p1, p2, p3

    def plot_integration_window(self, **kwargs):
        """
        Plots the three principal cross-sections of the integration volume on a single figure.

        Parameters
        ----------
        **kwargs : keyword arguments, optional
            Additional keyword arguments to customize the plot.
        """
        data = self.integration_volume
        center = self.center

        fig, axes = plt.subplots(1, 3, figsize=(15, 4))

        # Plot cross section 1
        slice_obj = [slice(None)] * data.ndim
        slice_obj[2] = center[2]
        p1 = plot_slice(data[slice_obj],
                        X=data[data.axes[0]],
                        Y=data[data.axes[1]],
                        ax=axes[0],
                        **kwargs)
        axes[0].set_aspect(len(data[data.axes[0]].nxdata) / len(data[data.axes[1]].nxdata))

        # Plot cross section 2
        slice_obj = [slice(None)] * data.ndim
        slice_obj[1] = center[1]
        p3 = plot_slice(data[slice_obj],
                        X=data[data.axes[0]],
                        Y=data[data.axes[2]],
                        ax=axes[1],
                        **kwargs)
        axes[1].set_aspect(len(data[data.axes[0]].nxdata) / len(data[data.axes[2]].nxdata))

        # Plot cross section 3
        slice_obj = [slice(None)] * data.ndim
        slice_obj[0] = center[0]
        p2 = plot_slice(data[slice_obj],
                        X=data[data.axes[1]],
                        Y=data[data.axes[2]],
                        ax=axes[2],
                        **kwargs)
        axes[2].set_aspect(len(data[data.axes[1]].nxdata) / len(data[data.axes[2]].nxdata))

        # Adjust subplot padding
        fig.subplots_adjust(wspace=0.3)

        plt.show()

        return p1, p2, p3


def reciprocal_lattice_params(lattice_params):
    """
    Calculate the reciprocal lattice parameters from the given direct lattice parameters.

    Parameters
    ----------
    lattice_params : tuple
        A tuple containing the direct lattice parameters (a, b, c, alpha, beta, gamma), where
        a, b, and c are the magnitudes of the lattice vectors, and alpha, beta, and gamma are the
        angles between them in degrees.

    Returns
    -------
    tuple
        A tuple containing the reciprocal lattice parameters (a*, b*, c*, alpha*, beta*, gamma*),
        where a*, b*, and c* are the magnitudes of the reciprocal lattice vectors, and alpha*,
        beta*, and gamma* are the angles between them in degrees.
    """
    a_mag, b_mag, c_mag, alpha, beta, gamma = lattice_params
    # Convert angles to radians
    alpha = np.deg2rad(alpha)
    beta = np.deg2rad(beta)
    gamma = np.deg2rad(gamma)

    # Calculate unit cell volume
    V = a_mag * b_mag * c_mag * np.sqrt(
        1 - np.cos(alpha) ** 2 - np.cos(beta) ** 2 - np.cos(gamma) ** 2
        + 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma)
    )

    # Calculate reciprocal lattice parameters
    a_star = (b_mag * c_mag * np.sin(alpha)) / V
    b_star = (a_mag * c_mag * np.sin(beta)) / V
    c_star = (a_mag * b_mag * np.sin(gamma)) / V
    alpha_star = np.rad2deg(np.arccos((np.cos(beta) * np.cos(gamma) - np.cos(alpha))
                                      / (np.sin(beta) * np.sin(gamma))))
    beta_star = np.rad2deg(np.arccos((np.cos(alpha) * np.cos(gamma) - np.cos(beta))
                                     / (np.sin(alpha) * np.sin(gamma))))
    gamma_star = np.rad2deg(np.arccos((np.cos(alpha) * np.cos(beta) - np.cos(gamma))
                                      / (np.sin(alpha) * np.sin(beta))))

    return a_star, b_star, c_star, alpha_star, beta_star, gamma_star


def rotate_data(data, lattice_angle, rotation_angle, rotation_axis, printout=False):
    """
    Rotates 3D data around a specified axis.

    Parameters
    ----------
    data : :class:`nexusformat.nexus.NXdata`
        Input data.
    lattice_angle : float
        Angle between the two in-plane lattice axes in degrees.
    rotation_angle : float
        Angle of rotation in degrees.
    rotation_axis : int
        Axis of rotation (0, 1, or 2).
    printout : bool, optional
        Enables printout of rotation progress. If set to True, information
        about each rotation slice will be printed to the console, indicating
        the axis being rotated and the corresponding coordinate value.
        Defaults to False.


    Returns
    -------
    rotated_data : :class:`nexusformat.nexus.NXdata`
        Rotated data as an NXdata object.
    """
    # Define output array
    output_array = np.zeros(data[data.signal].shape)

    # Define transformation
    skew_angle_adj = 90 - lattice_angle
    t = Affine2D()
    # Scale y-axis to preserve norm while shearing
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180))
    # Shear along x-axis
    t += Affine2D().skew_deg(skew_angle_adj, 0)
    # Return to original y-axis scaling
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180)).inverted()

    for i in range(len(data[data.axes[rotation_axis]])):
        if printout:
            print(f'\rRotating {data.axes[rotation_axis]}'
                  f'={data[data.axes[rotation_axis]][i]}...                      ',
                  end='', flush=True)
        # Identify current slice
        if rotation_axis == 0:
            sliced_data = data[i, :, :]
        elif rotation_axis == 1:
            sliced_data = data[:, i, :]
        elif rotation_axis == 2:
            sliced_data = data[:, :, i]
        else:
            sliced_data = None

        p = Padder(sliced_data)
        padding = tuple(len(sliced_data[axis]) for axis in sliced_data.axes)
        counts = p.pad(padding).counts

        counts_skewed = ndimage.affine_transform(counts,
                                                 t.inverted().get_matrix()[:2, :2],
                                                 offset=[counts.shape[0] / 2
                                                         * np.sin(skew_angle_adj * np.pi / 180),
                                                         0],
                                                 order=0,
                                                 )
        scale1 = np.cos(skew_angle_adj * np.pi / 180)
        counts_scaled1 = ndimage.affine_transform(counts_skewed,
                                                  Affine2D().scale(scale1, 1).get_matrix()[:2, :2],
                                                  offset=[(1 - scale1) * counts.shape[0] / 2, 0],
                                                  order=0,
                                                  )
        scale2 = counts.shape[0] / counts.shape[1]
        counts_scaled2 = ndimage.affine_transform(counts_scaled1,
                                                  Affine2D().scale(scale2, 1).get_matrix()[:2, :2],
                                                  offset=[(1 - scale2) * counts.shape[0] / 2, 0],
                                                  order=0,
                                                  )

        counts_rotated = ndimage.rotate(counts_scaled2, rotation_angle, reshape=False, order=0)

        counts_unscaled2 = ndimage.affine_transform(counts_rotated,
                                                    Affine2D().scale(
                                                        scale2, 1
                                                    ).inverted().get_matrix()[:2, :2],
                                                    offset=[-(1 - scale2) * counts.shape[
                                                        0] / 2 / scale2, 0],
                                                    order=0,
                                                    )

        counts_unscaled1 = ndimage.affine_transform(counts_unscaled2,
                                                    Affine2D().scale(
                                                        scale1, 1
                                                    ).inverted().get_matrix()[:2, :2],
                                                    offset=[-(1 - scale1) * counts.shape[
                                                        0] / 2 / scale1, 0],
                                                    order=0,
                                                    )

        counts_unskewed = ndimage.affine_transform(counts_unscaled1,
                                                   t.get_matrix()[:2, :2],
                                                   offset=[
                                                       (-counts.shape[0] / 2
                                                        * np.sin(skew_angle_adj * np.pi / 180)),
                                                       0],
                                                   order=0,
                                                   )

        counts_unpadded = p.unpad(counts_unskewed)

        # Write current slice
        if rotation_axis == 0:
            output_array[i, :, :] = counts_unpadded
        elif rotation_axis == 1:
            output_array[:, i, :] = counts_unpadded
        elif rotation_axis == 2:
            output_array[:, :, i] = counts_unpadded
    print('\nDone.')
    return NXdata(NXfield(output_array, name='counts'),
                  (data[data.axes[0]], data[data.axes[1]], data[data.axes[2]]))


def rotate_data2D(data, lattice_angle, rotation_angle):
    """
    Rotates 2D data.

    Parameters
    ----------
    data : :class:`nexusformat.nexus.NXdata`
        Input data.
    lattice_angle : float
        Angle between the two in-plane lattice axes in degrees.
    rotation_angle : float
        Angle of rotation in degrees.


    Returns
    -------
    rotated_data : :class:`nexusformat.nexus.NXdata`
        Rotated data as an NXdata object.
    """

    # Define transformation
    skew_angle_adj = 90 - lattice_angle
    t = Affine2D()
    # Scale y-axis to preserve norm while shearing
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180))
    # Shear along x-axis
    t += Affine2D().skew_deg(skew_angle_adj, 0)
    # Return to original y-axis scaling
    t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180)).inverted()

    p = Padder(data)
    padding = tuple(len(data[axis]) for axis in data.axes)
    counts = p.pad(padding).counts

    counts_skewed = ndimage.affine_transform(counts,
                                             t.inverted().get_matrix()[:2, :2],
                                             offset=[counts.shape[0] / 2
                                                     * np.sin(skew_angle_adj * np.pi / 180), 0],
                                             order=0,
                                             )
    scale1 = np.cos(skew_angle_adj * np.pi / 180)
    counts_scaled1 = ndimage.affine_transform(counts_skewed,
                                              Affine2D().scale(scale1, 1).get_matrix()[:2, :2],
                                              offset=[(1 - scale1) * counts.shape[0] / 2, 0],
                                              order=0,
                                              )
    scale2 = counts.shape[0] / counts.shape[1]
    counts_scaled2 = ndimage.affine_transform(counts_scaled1,
                                              Affine2D().scale(scale2, 1).get_matrix()[:2, :2],
                                              offset=[(1 - scale2) * counts.shape[0] / 2, 0],
                                              order=0,
                                              )

    counts_rotated = ndimage.rotate(counts_scaled2, rotation_angle, reshape=False, order=0)

    counts_unscaled2 = ndimage.affine_transform(counts_rotated,
                                                Affine2D().scale(
                                                    scale2, 1
                                                ).inverted().get_matrix()[:2, :2],
                                                offset=[-(1 - scale2) * counts.shape[
                                                    0] / 2 / scale2, 0],
                                                order=0,
                                                )

    counts_unscaled1 = ndimage.affine_transform(counts_unscaled2,
                                                Affine2D().scale(
                                                    scale1, 1
                                                ).inverted().get_matrix()[:2, :2],
                                                offset=[-(1 - scale1) * counts.shape[
                                                    0] / 2 / scale1, 0],
                                                order=0,
                                                )

    counts_unskewed = ndimage.affine_transform(counts_unscaled1,
                                               t.get_matrix()[:2, :2],
                                               offset=[
                                                   (-counts.shape[0] / 2
                                                    * np.sin(skew_angle_adj * np.pi / 180)),
                                                   0],
                                               order=0,
                                               )

    counts_unpadded = p.unpad(counts_unskewed)

    print('\nDone.')
    return NXdata(NXfield(counts_unpadded, name='counts'),
                  (data[data.axes[0]], data[data.axes[1]]))


class Padder:
    """
    A class to symmetrically pad and unpad datasets with a region of zeros.

    Attributes
    ----------
    data : NXdata or None
        The input data to be padded.
    padded : NXdata or None
        The padded data with symmetric zero padding.
    padding : tuple or None
        The number of zero-value pixels added along each edge of the array.
    steps : tuple or None
        The step sizes along each axis of the dataset.
    maxes : tuple or None
        The maximum values along each axis of the dataset.

    Methods
    -------
    set_data(data)
        Set the input data for padding.
    pad(padding)
        Symmetrically pads the data with zero values.
    save(fout_name=None)
        Saves the padded dataset to a .nxs file.
    unpad(data)
        Removes the padded region from the data.
    """

    def __init__(self, data=None):
        """
        Initialize the Padder object.

        Parameters
        ----------
        data : NXdata, optional
            The input data to be padded. If provided, the `set_data` method
            is called to set the data.
        """
        self.padded = None
        self.padding = None
        if data is not None:
            self.set_data(data)

    def set_data(self, data):
        """
        Set the input data for padding.

        Parameters
        ----------
        data : NXdata
            The input data to be padded.
        """
        self.data = data

        self.steps = tuple((data[axis].nxdata[1] - data[axis].nxdata[0])
                            for axis in data.axes)

        # Absolute value of the maximum value; assumes the domain of the input
        # is symmetric (eg, -H_min = H_max)
        self.maxes = tuple(data[axis].nxdata.max() for axis in data.axes)

    def pad(self, padding):
        """
        Symmetrically pads the data with zero values.

        Parameters
        ----------
        padding : tuple
            The number of zero-value pixels to add along each edge of the array.

        Returns
        -------
        NXdata
            The padded data with symmetric zero padding.
        """
        data = self.data
        self.padding = padding

        padded_shape = tuple(data[data.signal].nxdata.shape[i]
                             + self.padding[i] * 2 for i in range(data.ndim))

        # Create padded dataset
        padded = np.zeros(padded_shape)

        slice_obj = [slice(None)] * data.ndim
        for i, _ in enumerate(slice_obj):
            slice_obj[i] = slice(self.padding[i], -self.padding[i], None)
        slice_obj = tuple(slice_obj)
        padded[slice_obj] = data[data.signal].nxdata

        padmaxes = tuple(self.maxes[i] + self.padding[i] * self.steps[i]
                         for i in range(data.ndim))

        padded = NXdata(NXfield(padded, name=data.signal),
                        tuple(NXfield(np.linspace(-padmaxes[i], padmaxes[i], padded_shape[i]),
                                      name=data.axes[i])
                              for i in range(data.ndim)))

        self.padded = padded
        return padded

    def save(self, fout_name=None):
        """
        Saves the padded dataset to a .nxs file.

        Parameters
        ----------
        fout_name : str, optional
            The output file name. Default is padded_(Hpadding)_(Kpadding)_(Lpadding).nxs
        """
        padH, padK, padL = self.padding

        # Save padded dataset
        print("Saving padded dataset...")
        f = NXroot()
        f['entry'] = NXentry()
        f['entry']['data'] = self.padded
        if fout_name is None:
            fout_name = 'padded_' + str(padH) + '_' + str(padK) + '_' + str(padL) + '.nxs'
        nxsave(fout_name, f)
        print("Output file saved to: " + os.path.join(os.getcwd(), fout_name))

    def unpad(self, data):
        """
        Removes the padded region from the data.

        Parameters
        ----------
        data : ndarray or NXdata
            The padded data from which to remove the padding.

        Returns
        -------
        ndarray or NXdata
            The unpadded data, with the symmetric padding region removed.
        """
        slice_obj = [slice(None)] * data.ndim
        for i in range(data.ndim):
            slice_obj[i] = slice(self.padding[i], -self.padding[i], None)
        slice_obj = tuple(slice_obj)
        return data[slice_obj]


def load_discus_nxs(path):
    """
    Load .nxs format data from the DISCUS program (by T. Proffen and R. Neder)
    and convert it to the CHESS format.

    Parameters
    ----------
    path : str
        The file path to the .nxs file generated by DISCUS.

    Returns
    -------
    NXdata
        The data converted to the CHESS format, with axes labeled 'H', 'K', and 'L',
        and the signal labeled 'counts'.

    """
    filename = path
    root = nxload(filename)
    hlim, klim, llim = root.lower_limits
    hstep, kstep, lstep = root.step_sizes
    h = NXfield(np.linspace(hlim, -hlim, int(np.abs(hlim * 2) / hstep) + 1), name='H')
    k = NXfield(np.linspace(klim, -klim, int(np.abs(klim * 2) / kstep) + 1), name='K')
    l = NXfield(np.linspace(llim, -llim, int(np.abs(llim * 2) / lstep) + 1), name='L')
    data = NXdata(NXfield(root.data[:, :, :], name='counts'), (h, k, l))

    return data



"""
Module for fitting of linecuts using the lmfit package.
"""

import operator
from lmfit.model import Model
from lmfit.model import CompositeModel
import matplotlib.pyplot as plt
import numpy as np


class LinecutModel:
    """
    A class representing a linecut model for data analysis and fitting.

    Attributes
    ----------
    y : array-like or None
        The dependent variable data.
    x : array-like or None
        The independent variable data.
    y_eval : array-like or None
        The evaluated y-values of the fitted model.
    x_eval : array-like or None
        The x-values used for evaluation.
    y_eval_components : dict or None
        The evaluated y-values of the model components.
    y_fit_components : dict or None
        The fitted y-values of the model components.
    y_fit : array-like or None
        The fitted y-values of the model.
    x_fit : array-like or None
        The x-values used for fitting.
    y_init_fit : array-like or None
        The initial guess of the y-values.
    params : Parameters or None
        The parameters of the model.
    model_components : Model or list of Models or None
        The model component(s) used for fitting.
    model : Model or None
        The composite model used for fitting.
    modelresult : ModelResult or None
        The result of the model fitting.
    data : NXdata or None
        The 1D linecut data used for analysis.

    Methods
    -------
    __init__(self, data=None)
        Initialize the LinecutModel.
    set_data(self, data)
        Set the data for analysis.
    set_model_components(self, model_components)
        Set the model components.
    set_param_hint(self, *args, **kwargs)
        Set parameter hints for the model.
    make_params(self)
        Create and initialize the parameters for the model.
    guess(self)
        Perform initial guesses for each model component.
    print_initial_params(self)
        Print out initial guesses for each parameter of the model.
    plot_initial_guess(self, numpoints=None)
        Plot the initial guess.
    fit(self)
        Fit the model to the data.
    plot_fit(self, numpoints=None, fit_report=True, **kwargs)
        Plot the fitted model.
    print_fit_report(self)
        Print the fit report.
    """
    def __init__(self, data=None):
        """
        Initialize the LinecutModel.
        """
        self.y = None
        self.x = None
        self.y_eval = None
        self.x_eval = None
        self.y_eval_components = None
        self.y_fit_components = None
        self.y_fit = None
        self.x_fit = None
        self.y_init_fit = None
        self.params = None
        self.model_components = None
        self.model = None
        self.modelresult = None
        self.data = data if data is not None else None
        if self.data is not None:
            self.x = data[data.axes].nxdata
            self.y = data[data.signal].nxdata

    def set_data(self, data):
        """
        Set the data for analysis.

        Parameters
        ----------
        data : NXdata
            The 1D linecut data to be used for analysis.
        """
        self.data = data
        self.x = data[data.axes].nxdata
        self.y = data[data.signal].nxdata

    def set_model_components(self, model_components):
        """
        Set the model components

        Parameters
        ----------
        model_components : Model or list of Models
            The model component(s) to be used for fitting,
             which will be combined into a CompositeModel.
        """

        # If the model only has one component, then use it as the model
        if isinstance(model_components, Model):
            self.model = model_components
        # Else, combine the components into a composite model and use that as the
        else:
            self.model_components = model_components
            self.model = model_components[0]

            # Combine remaining components into the composite model
            for component in model_components[1:]:
                self.model = CompositeModel(self.model, component, operator.add)

    def set_param_hint(self, *args, **kwargs):
        """
        Set parameter hints for the model.

        Parameters
        ----------
        *args : positional arguments
            Positional arguments passed to the `set_param_hint` method of the underlying model.

        **kwargs : keyword arguments
            Keyword arguments passed to the `set_param_hint` method of the underlying model.
        """

        self.model.set_param_hint(*args, **kwargs)

    def make_params(self):
        """
        Create and initialize the parameters for the model.

        Returns
        -------
        Parameters
            The initialized parameters for the model.
        """
        # Intialize empty parameters (in function)
        params = self.model.make_params()
        self.params = params

        return params

    def guess(self):
        """
        Perform initial guesses for each model component.
        """
        for model_component in list(self.model_components):
            self.params.update(model_component.guess(self.y, x=self.x))

    def print_initial_params(self):
        """
        Print out initial guesses for each parameter of the model.
        """
        model = self.model
        for param, hint in model.param_hints.items():
            print(f'{param}')
            for key, value in hint.items():
                print(f'\t{key}: {value}')

    def plot_initial_guess(self, numpoints=None):
        """
        Plot initial guess.
        """
        model = self.model
        params = self.params
        x = self.x
        y = self.y
        y_init_fit = model.eval(params=params, x=x)
        self.y_init_fit = y_init_fit
        plt.plot(x, y, 'o', label='data')
        plt.plot(x, y_init_fit, '--', label='guess')

        # Plot the components of the model
        if numpoints is None:
            numpoints = len(self.x)
        self.x_eval = np.linspace(self.x.min(), self.x.max(), numpoints)
        y_init_fit_components = model.eval_components(params=params, x=self.x_eval)
        ax = plt.gca()
        for model_component, value in y_init_fit_components.items():
            ax.fill_between(self.x_eval, value, alpha=0.3, label=model_component)
        plt.legend()
        plt.show()

    def fit(self):
        """
        Fit the model to the data.

        This method fits the model to the data using the specified parameters and the x-values.
        It updates the model result, fitted y-values, and the evaluated components.

        """
        self.modelresult = self.model.fit(self.y, self.params, x=self.x)
        self.y_fit = self.modelresult.eval(x=self.x)
        self.y_fit_components = self.modelresult.eval_components(x=self.x)

    def plot_fit(self, numpoints=None, fit_report=True, **kwargs):
        """
        Plot the fitted model.

        This method plots the fitted model along with the original data.
        It evaluates the model and its components at the specified number of points (numpoints)
        and plots the results.

        Parameters
        ----------
        numpoints : int, optional
            Number of points to evaluate the model and its components. If not provided,
            it defaults to the length of the x-values.

        fit_report : bool, optional
            Whether to print the fit report. Default is True.

        **kwargs : dict, optional
            Additional keyword arguments to be passed to the `plot` method.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The Axes object containing the plot.

        """
        if numpoints is None:
            numpoints = len(self.x)
        self.x_eval = np.linspace(self.x.min(), self.x.max(), numpoints)
        self.y_eval = self.modelresult.eval(x=self.x_eval)
        self.y_eval_components = self.modelresult.eval_components(x=self.x_eval)
        self.modelresult.plot(numpoints=numpoints, **kwargs)
        ax = plt.gca()
        for model_component, value in self.y_eval_components.items():
            ax.fill_between(self.x_eval, value, alpha=0.3, label=model_component)
            # ax.plot(self.x_eval, value, label=model_component)
        plt.legend()
        plt.show()
        if fit_report:
            print(self.modelresult.fit_report())
        return ax

    def print_fit_report(self):
        """
        Show fit report.
        """
        print(self.modelresult.fit_report())



"""
Tools for generating single crystal pair distribution functions.
"""
import time
import os
import gc
import math
from scipy import ndimage
import scipy
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D
from nexusformat.nexus import nxsave, NXroot, NXentry, NXdata, NXfield
import numpy as np
from astropy.convolution import Kernel, convolve_fft
import pyfftw
from .datareduction import plot_slice, reciprocal_lattice_params, Padder, \
    array_to_nxdata

__all__ = ['Symmetrizer2D', 'Symmetrizer3D', 'Puncher', 'Interpolator',
           'fourier_transform_nxdata', 'Gaussian3DKernel', 'DeltaPDF',
           'generate_gaussian'
           ]


class Symmetrizer2D:
    """
    A class for symmetrizing 2D datasets.

    The `Symmetrizer2D` class provides functionality to apply symmetry
    operations such as rotation and mirroring to 2D datasets.

    Attributes
    ----------
    mirror_axis : int or None
        The axis along which mirroring is performed. Default is None, meaning
        no mirroring is applied.
    symmetrized : NXdata or None
        The symmetrized dataset after applying the symmetrization operations.
        Default is None until symmetrization is performed.
    wedges : NXdata or None
        The wedges extracted from the dataset based on the angular limits.
        Default is None until symmetrization is performed.
    rotations : int or None
        The number of rotations needed to reconstruct the full dataset from
        a single wedge. Default is None until parameters are set.
    transform : Affine2D or None
        The transformation matrix used for skewing and scaling the dataset.
        Default is None until parameters are set.
    mirror : bool or None
        Indicates whether mirroring is performed during symmetrization.
        Default is None until parameters are set.
    skew_angle : float or None
        The skew angle (in degrees) between the principal axes of the plane
         to be symmetrized. Default is None until parameters are set.
    theta_max : float or None
        The maximum angle (in degrees) for symmetrization. Default is None
         until parameters are set.
    theta_min : float or None
        The minimum angle (in degrees) for symmetrization. Default is None
         until parameters are set.
    wedge : NXdata or None
        The dataset wedge used in the symmetrization process. Default is
         None until symmetrization is performed.
    symmetrization_mask : NXdata or None
        The mask used for selecting the region of the dataset to be symmetrized.
         Default is None until symmetrization is performed.

    Methods
    -------
    __init__(**kwargs):
        Initializes the Symmetrizer2D object and optionally sets the parameters
         using `set_parameters`.
    set_parameters(theta_min, theta_max, lattice_angle=90, mirror=True, mirror_axis=0):
        Sets the parameters for the symmetrization operation, including angle limits,
         lattice angle, and mirroring options.
    symmetrize_2d(data):
        Symmetrizes a 2D dataset based on the set parameters.
    test(data, **kwargs):
        Performs a test visualization of the symmetrization process, displaying the
         original data, mask, wedge, and symmetrized result.
    """
    symmetrization_mask: NXdata

    def __init__(self, **kwargs):
        """
        Initializes the Symmetrizer2D object.

        Parameters
        ----------
        **kwargs : dict, optional
            Keyword arguments that can be passed to the `set_parameters` method to
             set the symmetrization parameters during initialization.
        """
        self.mirror_axis = None
        self.symmetrized = None
        self.wedges = None
        self.rotations = None
        self.transform = None
        self.mirror = None
        self.skew_angle = None
        self.theta_max = None
        self.theta_min = None
        self.wedge = None
        if kwargs:
            self.set_parameters(**kwargs)

    def set_parameters(self, theta_min, theta_max, lattice_angle=90, mirror=True, mirror_axis=0):
        """
        Sets the parameters for the symmetrization operation, and calculates the
         required transformations and rotations.

        Parameters
        ----------
        theta_min : float
            The minimum angle in degrees for symmetrization.
        theta_max : float
            The maximum angle in degrees for symmetrization.
        lattice_angle : float, optional
            The angle in degrees between the two principal axes of the plane to be
             symmetrized (default: 90).
        mirror : bool, optional
            If True, perform mirroring during symmetrization (default: True).
        mirror_axis : int, optional
            The axis along which to perform mirroring (default: 0).
        """
        self.theta_min = theta_min
        self.theta_max = theta_max
        self.skew_angle = lattice_angle
        self.mirror = mirror
        self.mirror_axis = mirror_axis

        # Define Transformation
        skew_angle_adj = 90 - lattice_angle
        t = Affine2D()
        # Scale y-axis to preserve norm while shearing
        t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180))
        # Shear along x-axis
        t += Affine2D().skew_deg(skew_angle_adj, 0)
        # Return to original y-axis scaling
        t += Affine2D().scale(1, np.cos(skew_angle_adj * np.pi / 180)).inverted()
        self.transform = t

        # Calculate number of rotations needed to reconstruct the dataset
        if mirror:
            rotations = abs(int(360 / (theta_max - theta_min) / 2))
        else:
            rotations = abs(int(360 / (theta_max - theta_min)))
        self.rotations = rotations

        self.symmetrization_mask = None

        self.wedges = None

        self.symmetrized = None

    def symmetrize_2d(self, data):
        """
        Symmetrizes a 2D dataset based on the set parameters, applying padding
         to prevent rotation cutoff and handling overlapping pixels.

        Parameters
        ----------
        data : NXdata
            The input 2D dataset to be symmetrized.

        Returns
        -------
        symmetrized : NXdata
            The symmetrized 2D dataset.
        """
        theta_min = self.theta_min
        theta_max = self.theta_max
        mirror = self.mirror
        mirror_axis = self.mirror_axis
        t = self.transform
        rotations = self.rotations

        # Pad the dataset so that rotations don't get cutoff if they extend
        # past the extent of the dataset
        p = Padder(data)
        padding = tuple(len(data[axis]) for axis in data.axes)
        data_padded = p.pad(padding)

        # Define axes that span the plane to be transformed
        q1 = data_padded[data.axes[0]]
        q2 = data_padded[data.axes[1]]

        # Define signal to be symmetrized
        counts = data_padded[data.signal].nxdata

        # Calculate the angle for each data point
        theta = np.arctan2(q1.reshape((-1, 1)), q2.reshape((1, -1)))
        # Create a boolean array for the range of angles
        symmetrization_mask = np.logical_and(theta >= theta_min * np.pi / 180,
                                             theta <= theta_max * np.pi / 180)
        self.symmetrization_mask = NXdata(NXfield(p.unpad(symmetrization_mask),
                                                  name='mask'),
                                          (data[data.axes[0]], data[data.axes[1]])
                                          )

        self.wedge = NXdata(NXfield(p.unpad(counts * symmetrization_mask),
                                    name=data.signal),
                            (data[data.axes[0]], data[data.axes[1]])
                            )

        # Scale and skew counts
        skew_angle_adj = 90 - self.skew_angle
        counts_skew = ndimage.affine_transform(counts,
                                               t.inverted().get_matrix()[:2, :2],
                                               offset=[counts.shape[0] / 2
                                                       * np.sin(skew_angle_adj * np.pi / 180), 0],
                                               order=0,
                                               )
        scale1 = np.cos(skew_angle_adj * np.pi / 180)
        wedge = ndimage.affine_transform(counts_skew,
                                         Affine2D().scale(scale1, 1).get_matrix()[:2, :2],
                                         offset=[(1 - scale1) * counts.shape[0] / 2, 0],
                                         order=0,
                                         ) * symmetrization_mask

        scale2 = counts.shape[0] / counts.shape[1]
        wedge = ndimage.affine_transform(wedge,
                                         Affine2D().scale(scale2, 1).get_matrix()[:2, :2],
                                         offset=[(1 - scale2) * counts.shape[0] / 2, 0],
                                         order=0,
                                         )

        # Reconstruct full dataset from wedge
        reconstructed = np.zeros(counts.shape)
        for _ in range(0, rotations):
            # The following are attempts to combine images with minimal overlapping pixels
            reconstructed += wedge
            # reconstructed = np.where(reconstructed == 0, reconstructed + wedge, reconstructed)

            wedge = ndimage.rotate(wedge, 360 / rotations, reshape=False, order=0)

        # self.rotated_only = NXdata(NXfield(reconstructed, name=data.signal),
        #                            (q1, q2))

        if mirror:
            # The following are attempts to combine images with minimal overlapping pixels
            reconstructed = np.where(reconstructed == 0,
                                     reconstructed + np.flip(reconstructed, axis=mirror_axis),
                                     reconstructed)
            # reconstructed += np.flip(reconstructed, axis=0)

        # self.rotated_and_mirrored = NXdata(NXfield(reconstructed, name=data.signal),
        #                                    (q1, q2))

        reconstructed = ndimage.affine_transform(reconstructed,
                                                 Affine2D().scale(
                                                     scale2, 1
                                                 ).inverted().get_matrix()[:2, :2],
                                                 offset=[-(1 - scale2) * counts.shape[
                                                     0] / 2 / scale2, 0],
                                                 order=0,
                                                 )
        reconstructed = ndimage.affine_transform(reconstructed,
                                                 Affine2D().scale(
                                                     scale1, 1
                                                 ).inverted().get_matrix()[:2, :2],
                                                 offset=[-(1 - scale1) * counts.shape[
                                                     0] / 2 / scale1, 0],
                                                 order=0,
                                                 )
        reconstructed = ndimage.affine_transform(reconstructed,
                                                 t.get_matrix()[:2, :2],
                                                 offset=[(-counts.shape[0] / 2
                                                          * np.sin(skew_angle_adj * np.pi / 180)),
                                                         0],
                                                 order=0,
                                                 )

        reconstructed_unpadded = p.unpad(reconstructed)

        # Fix any overlapping pixels by truncating counts to max
        reconstructed_unpadded[reconstructed_unpadded > data[data.signal].nxdata.max()] \
            = data[data.signal].nxdata.max()

        symmetrized = NXdata(NXfield(reconstructed_unpadded, name=data.signal),
                             (data[data.axes[0]],
                              data[data.axes[1]]))

        return symmetrized

    def test(self, data, **kwargs):
        """
        Performs a test visualization of the symmetrization process to help assess
         the effect of the parameters.

        Parameters
        ----------
        data : ndarray
            The input 2D dataset to be used for the test visualization.
        **kwargs : dict
            Additional keyword arguments to be passed to the plot_slice function.

        Returns
        -------
        fig : Figure
            The matplotlib Figure object that contains the test visualization plot.
        axesarr : ndarray
            The numpy array of Axes objects representing the subplots in the test
             visualization.

        Notes
        -----
        This method uses the `symmetrize_2d` method to perform the symmetrization on
         the input data and visualize the process.

        The test visualization plot includes the following subplots:
        - Subplot 1: The original dataset.
        - Subplot 2: The symmetrization mask.
        - Subplot 3: The wedge slice used for reconstruction of the full symmetrized dataset.
        - Subplot 4: The symmetrized dataset.

        Example usage:
        ```
        s = Symmetrizer2D()
        s.set_parameters(theta_min, theta_max, skew_angle, mirror)
        s.test(data)
        ```
        """
        s = self
        symm_test = s.symmetrize_2d(data)
        fig, axesarr = plt.subplots(2, 2, figsize=(10, 8))
        axes = axesarr.reshape(-1)
        plot_slice(data, skew_angle=s.skew_angle, ax=axes[0], title='data', **kwargs)
        plot_slice(s.symmetrization_mask, skew_angle=s.skew_angle, ax=axes[1], title='mask')
        plot_slice(s.wedge, skew_angle=s.skew_angle, ax=axes[2], title='wedge', **kwargs)
        plot_slice(symm_test, skew_angle=s.skew_angle, ax=axes[3], title='symmetrized', **kwargs)
        plt.subplots_adjust(wspace=0.4)
        plt.show()
        return fig, axesarr


class Symmetrizer3D:
    """
    A class to symmetrize 3D datasets by performing sequential 2D symmetrization on
     different planes.

    This class applies 2D symmetrization on the three principal planes of a 3D dataset,
    effectively enhancing the symmetry of the data across all axes.
    """

    def __init__(self, data=None):
        """
        Initialize the Symmetrizer3D object with an optional 3D dataset.

        If data is provided, the corresponding q-vectors and planes are automatically
         set up for symmetrization.

        Parameters
        ----------
        data : NXdata, optional
            The input 3D dataset to be symmetrized.
        """

        self.a, self.b, self.c, self.al, self.be, self.ga = [None] * 6
        self.a_star, self.b_star, self.c_star, self.al_star, self.be_star, self.ga_star = [None] * 6
        self.lattice_params = None
        self.reciprocal_lattice_params = None
        self.symmetrized = None
        self.data = data
        self.plane1symmetrizer = Symmetrizer2D()
        self.plane2symmetrizer = Symmetrizer2D()
        self.plane3symmetrizer = Symmetrizer2D()

        if data is not None:
            self.q1 = data[data.axes[0]]
            self.q2 = data[data.axes[1]]
            self.q3 = data[data.axes[2]]
            self.plane1 = self.q1.nxname + self.q2.nxname
            self.plane2 = self.q1.nxname + self.q3.nxname
            self.plane3 = self.q2.nxname + self.q3.nxname

        print("Plane 1: " + self.plane1)
        print("Plane 2: " + self.plane2)
        print("Plane 3: " + self.plane3)

    def set_data(self, data):
        """
        Sets the 3D dataset to be symmetrized and updates the corresponding q-vectors and planes.

        Parameters
        ----------
        data : NXdata
            The input 3D dataset to be symmetrized.
        """
        self.data = data
        self.q1 = data[data.axes[0]]
        self.q2 = data[data.axes[1]]
        self.q3 = data[data.axes[2]]
        self.plane1 = self.q1.nxname + self.q2.nxname
        self.plane2 = self.q1.nxname + self.q3.nxname
        self.plane3 = self.q2.nxname + self.q3.nxname

        print("Plane 1: " + self.plane1)
        print("Plane 2: " + self.plane2)
        print("Plane 3: " + self.plane3)

    def set_lattice_params(self, lattice_params):
        """
        Sets the lattice parameters and calculates the reciprocal lattice parameters.

        Parameters
        ----------
        lattice_params : tuple of float
            The lattice parameters (a, b, c, alpha, beta, gamma) in real space.
        """
        self.a, self.b, self.c, self.al, self.be, self.ga = lattice_params
        self.lattice_params = lattice_params
        self.reciprocal_lattice_params = reciprocal_lattice_params(lattice_params)
        self.a_star, self.b_star, self.c_star, \
            self.al_star, self.be_star, self.ga_star = self.reciprocal_lattice_params

    def symmetrize(self):
        """
        Perform the symmetrization of the 3D dataset by sequentially applying
         2D symmetrization on the three principal planes.

        This method symmetrizes the dataset on the three principal planes
         (q1-q2, q1-q3, q2-q3) and handles any negative values that might result
          from the symmetrization process.

        Returns
        -------
        symmetrized : NXdata
            The symmetrized 3D dataset.
        """
        starttime = time.time()
        data = self.data
        q1, q2, q3 = self.q1, self.q2, self.q3
        out_array = np.zeros(data[data.signal].shape)

        if self.plane1symmetrizer.theta_max:
            print('Symmetrizing ' + self.plane1 + ' planes...')
            for k, value in enumerate(q3):
                print(f'Symmetrizing {q3.nxname}={value:.02f}...', end='\r')
                data_symmetrized = self.plane1symmetrizer.symmetrize_2d(data[:, :, k])
                out_array[:, :, k] = data_symmetrized[data.signal].nxdata
            print('\nSymmetrized ' + self.plane1 + ' planes.')

        if self.plane2symmetrizer.theta_max:
            print('Symmetrizing ' + self.plane2 + ' planes...')
            for j, value in enumerate(q2):
                print(f'Symmetrizing {q2.nxname}={value:.02f}...', end='\r')
                data_symmetrized = self.plane2symmetrizer.symmetrize_2d(
                    NXdata(NXfield(out_array[:, j, :], name=data.signal), (q1, q3))
                )
                out_array[:, j, :] = data_symmetrized[data.signal].nxdata
            print('\nSymmetrized ' + self.plane2 + ' planes.')

        if self.plane3symmetrizer.theta_max:
            print('Symmetrizing ' + self.plane3 + ' planes...')
            for i, value in enumerate(q1):
                print(f'Symmetrizing {q1.nxname}={value:.02f}...', end='\r')
                data_symmetrized = self.plane3symmetrizer.symmetrize_2d(
                    NXdata(NXfield(out_array[i, :, :], name=data.signal), (q2, q3))
                )
                out_array[i, :, :] = data_symmetrized[data.signal].nxdata
            print('\nSymmetrized ' + self.plane3 + ' planes.')

        out_array[out_array < 0] = 0

        stoptime = time.time()
        print(f"\nSymmetrization finished in {((stoptime - starttime) / 60):.02f} minutes.")

        self.symmetrized = NXdata(NXfield(out_array, name=data.signal),
                                  tuple(data[axis] for axis in data.axes))

        return self.symmetrized

    def save(self, fout_name=None):
        """
        Save the symmetrized dataset to a NeXus file.

        Parameters
        ----------
        fout_name : str, optional
            The name of the output file. If not provided,
            the default name 'symmetrized.nxs' will be used.
        """
        print("Saving file...")

        f = NXroot()
        f['entry'] = NXentry()
        f['entry']['data'] = self.symmetrized
        if fout_name is None:
            fout_name = 'symmetrized.nxs'
        nxsave(fout_name, f)
        print("Output file saved to: " + os.path.join(os.getcwd(), fout_name))


def generate_gaussian(H, K, L, amp, stddev, lattice_params, coeffs=None):
    """
    Generate a 3D Gaussian distribution.

    This function creates a 3D Gaussian distribution in reciprocal space based
     on the specified parameters.

    Parameters
    ----------
    H, K, L : ndarray
        Arrays specifying the values of H, K, and L coordinates in reciprocal space.
    amp : float
        Amplitude of the Gaussian distribution.
    stddev : float
        Standard deviation of the Gaussian distribution.
    lattice_params : tuple
        Tuple of lattice parameters (a, b, c, alpha, beta, gamma) for the
         reciprocal lattice.
    coeffs : list, optional
        Coefficients for the Gaussian expression, including cross-terms between axes.
         Default is [1, 0, 1, 0, 1, 0],
         corresponding to (1*H**2 + 0*H*K + 1*K**2 + 0*K*L + 1*L**2 + 0*L*H).

    Returns
    -------
    gaussian : ndarray
        3D Gaussian distribution array.
    """
    if coeffs is None:
        coeffs = [1, 0, 1, 0, 1, 0]
    a, b, c, al, be, ga = lattice_params
    a_, b_, c_, _, _, _ = reciprocal_lattice_params((a, b, c, al, be, ga))
    H, K, L = np.meshgrid(H, K, L, indexing='ij')
    gaussian = amp * np.exp(-(coeffs[0] * H ** 2 +
                              coeffs[1] * (b_ * a_ / (a_ ** 2)) * H * K +
                              coeffs[2] * (b_ / a_) ** 2 * K ** 2 +
                              coeffs[3] * (b_ * c_ / (a_ ** 2)) * K * L +
                              coeffs[4] * (c_ / a_) ** 2 * L ** 2 +
                              coeffs[5] * (c_ * a_ / (a_ ** 2)) * L * H) / (2 * stddev ** 2))
    if gaussian.ndim == 3:
        gaussian = gaussian.transpose(1, 0, 2)
    elif gaussian.ndim == 2:
        gaussian = gaussian.transpose()
    return gaussian.transpose(1, 0, 2)


class Puncher:
    """
    A class for applying masks to 3D datasets, typically for data processing in reciprocal space.

    This class provides methods for setting data, applying masks, and generating
     masks based on various criteria. It can be used to "punch" or modify datasets
      by setting specific regions to NaN according to the mask.

    Attributes
    ----------
    punched : NXdata, optional
        The dataset with regions modified (punched) based on the mask.
    data : NXdata, optional
        The input dataset to be processed.
    HH, KK, LL : ndarray
        Meshgrid arrays representing the H, K, and L coordinates in reciprocal space.
    mask : ndarray, optional
        The mask used for identifying and modifying specific regions in the dataset.
    reciprocal_lattice_params : tuple, optional
        The reciprocal lattice parameters derived from the lattice parameters.
    lattice_params : tuple, optional
        The lattice parameters (a, b, c, alpha, beta, gamma).
    a, b, c, al, be, ga : float
        Individual lattice parameters.
    a_star, b_star, c_star, al_star, be_star, ga_star : float
        Individual reciprocal lattice parameters.

    Methods
    -------
    set_data(data)
        Sets the dataset to be processed and initializes the coordinate arrays and mask.
    set_lattice_params(lattice_params)
        Sets the lattice parameters and computes the reciprocal lattice parameters.
    add_mask(maskaddition)
        Adds regions to the current mask using a logical OR operation.
    subtract_mask(masksubtraction)
        Removes regions from the current mask using a logical AND NOT operation.
    generate_bragg_mask(punch_radius, coeffs=None, thresh=None)
        Generates a mask for Bragg peaks based on a Gaussian distribution in
         reciprocal space.
    generate_intensity_mask(thresh, radius, verbose=True)
        Generates a mask based on intensity thresholds, including a spherical
         region around high-intensity points.
    generate_mask_at_coord(coordinate, punch_radius, coeffs=None, thresh=None)
        Generates a mask centered at a specific coordinate in reciprocal space
         with a specified radius.
    punch()
        Applies the mask to the dataset, setting masked regions to NaN.
    """

    def __init__(self):
        """
        Initialize the Puncher object.

        This method sets up the initial state of the Puncher instance, including
         attributes for storing the dataset, lattice parameters, and masks.
          It prepares the object for further data processing and masking operations.

        Attributes
        ----------
        punched : NXdata, optional
            The dataset with modified (punched) regions, initialized as None.
        data : NXdata, optional
            The input dataset to be processed, initialized as None.
        HH, KK, LL : ndarray, optional
            Arrays representing the H, K, and L coordinates in reciprocal space,
             initialized as None.
        mask : ndarray, optional
            The mask for identifying and modifying specific regions in the dataset,
             initialized as None.
        reciprocal_lattice_params : tuple, optional
            The reciprocal lattice parameters, initialized as None.
        lattice_params : tuple, optional
            The lattice parameters (a, b, c, alpha, beta, gamma),
             initialized as None.
        a, b, c, al, be, ga : float
            Individual lattice parameters, initialized as None.
        a_star, b_star, c_star, al_star, be_star, ga_star : float
            Individual reciprocal lattice parameters, initialized as None.
        """
        self.punched = None
        self.data = None
        self.HH, self.KK, self.LL = [None] * 3
        self.mask = None
        self.reciprocal_lattice_params = None
        self.lattice_params = None
        self.a, self.b, self.c, self.al, self.be, self.ga = [None] * 6
        self.a_star, self.b_star, self.c_star, self.al_star, self.be_star, self.ga_star = [None] * 6

    def set_data(self, data):
        """
        Set the 3D dataset and initialize the mask if not already set.

        Parameters
        ----------
        data : NXdata
            The dataset to be processed.

        Notes
        -----
        This method also sets up the H, K, and L coordinate grids for the dataset.
        """
        self.data = data
        if self.mask is None:
            self.mask = np.zeros(data[data.signal].nxdata.shape)
        self.HH, self.KK, self.LL = np.meshgrid(data[data.axes[0]],
                                                data[data.axes[1]],
                                                data[data.axes[2]],
                                                indexing='ij')

    def set_lattice_params(self, lattice_params):
        """
        Set the lattice parameters and compute the reciprocal lattice parameters.

        Parameters
        ----------
        lattice_params : tuple
            Tuple of lattice parameters (a, b, c, alpha, beta, gamma).
        """
        self.a, self.b, self.c, self.al, self.be, self.ga = lattice_params
        self.lattice_params = lattice_params
        self.reciprocal_lattice_params = reciprocal_lattice_params(lattice_params)
        self.a_star, self.b_star, self.c_star, \
            self.al_star, self.be_star, self.ga_star = self.reciprocal_lattice_params

    def add_mask(self, maskaddition):
        """
        Add regions to the current mask using a logical OR operation.

        Parameters
        ----------
        maskaddition : ndarray
            The mask to be added.
        """
        self.mask = np.logical_or(self.mask, maskaddition)

    def subtract_mask(self, masksubtraction):
        """
        Remove regions from the current mask using a logical AND NOT operation.

        Parameters
        ----------
        masksubtraction : ndarray
            The mask to be subtracted.
        """
        self.mask = np.logical_and(self.mask, np.logical_not(masksubtraction))

    def generate_bragg_mask(self, punch_radius, coeffs=None, thresh=None):
        """
        Generate a mask for Bragg peaks.

        Parameters
        ----------
        punch_radius : float
            Radius for the Bragg peak mask.
        coeffs : list, optional
            Coefficients for the expression of the sphere to be removed around
            each Bragg position, corresponding to coefficients for H, HK, K, KL, L, and LH terms.
            Default is [1, 0, 1, 0, 1, 0].
        thresh : float, optional
            Intensity threshold for applying the mask.

        Returns
        -------
        mask : ndarray
            Boolean mask identifying the Bragg peaks.
        """
        if coeffs is None:
            coeffs = [1, 0, 1, 0, 1, 0]
        data = self.data
        H, K, L = self.HH, self.KK, self.LL
        a_, b_, c_, _, _, _ = self.reciprocal_lattice_params

        mask = (coeffs[0] * (H - np.rint(H)) ** 2 +
                coeffs[1] * (b_ * a_ / (a_ ** 2)) * (H - np.rint(H)) * (K - np.rint(K)) +
                coeffs[2] * (b_ / a_) ** 2 * (K - np.rint(K)) ** 2 +
                coeffs[3] * (b_ * c_ / (a_ ** 2)) * (K - np.rint(K)) * (L - np.rint(L)) +
                coeffs[4] * (c_ / a_) ** 2 * (L - np.rint(L)) ** 2 +
                coeffs[5] * (c_ * a_ / (a_ ** 2)) * (L - np.rint(L)) * (H - np.rint(H))) \
               < punch_radius ** 2

        if thresh:
            mask = np.logical_and(mask, data[data.signal] > thresh)

        return mask

    def generate_intensity_mask(self, thresh, radius, verbose=True):
        """
        Generate a mask based on intensity thresholds.

        Parameters
        ----------
        thresh : float
            Intensity threshold for creating the mask.
        radius : int
            Radius around high-intensity points to include in the mask.
        verbose : bool, optional
            Whether to print progress information.

        Returns
        -------
        mask : ndarray
            Boolean mask highlighting regions with high intensity.
        """
        data = self.data
        counts = data[data.signal].nxdata
        mask = np.zeros(counts.shape)

        print(f"Shape of data is {counts.shape}") if verbose else None
        for i in range(counts.shape[0]):
            for j in range(counts.shape[1]):
                for k in range(counts.shape[2]):
                    if counts[i, j, k] > thresh:
                        # Set the pixels within the sphere to NaN
                        for x in range(max(i - radius, 0),
                                       min(i + radius + 1, counts.shape[0])):
                            for y in range(max(j - radius, 0),
                                           min(j + radius + 1, counts.shape[1])):
                                for z in range(max(k - radius, 0),
                                               min(k + radius + 1, counts.shape[2])):
                                    mask[x, y, z] = 1
                        print(f"Found high intensity at ({i}, {j}, {k}).\t\t", end='\r') \
                            if verbose else None
        print("\nDone.")
        return mask

    def generate_mask_at_coord(self, coordinate, punch_radius, coeffs=None, thresh=None):
        """
        Generate a mask centered at a specific coordinate.

        Parameters
        ----------
        coordinate : tuple of float
            Center coordinate (H, K, L) for the mask.
        punch_radius : float
            Radius for the mask.
        coeffs : list, optional
            Coefficients for the expression of the sphere to be removed around
            each Bragg position,
            corresponding to coefficients for H, HK, K, KL, L, and LH terms.
            Default is [1, 0, 1, 0, 1, 0].
        thresh : float, optional
            Intensity threshold for applying the mask.

        Returns
        -------
        mask : ndarray
            Boolean mask for the specified coordinate.
        """
        if coeffs is None:
            coeffs = [1, 0, 1, 0, 1, 0]
        data = self.data
        H, K, L = self.HH, self.KK, self.LL
        a_, b_, c_, _, _, _ = self.reciprocal_lattice_params
        centerH, centerK, centerL = coordinate
        mask = (coeffs[0] * (H - centerH) ** 2 +
                coeffs[1] * (b_ * a_ / (a_ ** 2)) * (H - centerH) * (K - centerK) +
                coeffs[2] * (b_ / a_) ** 2 * (K - centerK) ** 2 +
                coeffs[3] * (b_ * c_ / (a_ ** 2)) * (K - centerK) * (L - centerL) +
                coeffs[4] * (c_ / a_) ** 2 * (L - centerL) ** 2 +
                coeffs[5] * (c_ * a_ / (a_ ** 2)) * (L - centerL) * (H - centerH)) \
               < punch_radius ** 2

        if thresh:
            mask = np.logical_and(mask, data[data.signal] > thresh)

        return mask

    def punch(self):
        """
        Apply the mask to the dataset, setting masked regions to NaN.

        This method creates a new dataset where the masked regions are set to
         NaN, effectively "punching" those regions.
        """
        data = self.data
        self.punched = NXdata(NXfield(
            np.where(self.mask, np.nan, data[data.signal].nxdata),
            name=data.signal),
            (data[data.axes[0]], data[data.axes[1]], data[data.axes[2]])
        )


def _round_up_to_odd_integer(value):
    """
    Round up a given number to the nearest odd integer.

    This function takes a floating-point value and rounds it up to the smallest
     odd integer that is greater than or equal to the given value.

    Parameters
    ----------
    value : float
        The input floating-point number to be rounded up.

    Returns
    -------
    int
        The nearest odd integer greater than or equal to the input value.

    Examples
    --------
    >>> _round_up_to_odd_integer(4.2)
    5

    >>> _round_up_to_odd_integer(5.0)
    5

    >>> _round_up_to_odd_integer(6.7)
    7
    """
    i = int(math.ceil(value))
    if i % 2 == 0:
        return i + 1

    return i


class Gaussian3DKernel(Kernel):
    """
    Initialize a 3D Gaussian kernel.

    This constructor creates a 3D Gaussian kernel with the specified
    standard deviation and size. The Gaussian kernel is generated based on
    the provided coefficients and is then normalized.

    Parameters
    ----------
    stddev : float
        The standard deviation of the Gaussian distribution, which controls
        the width of the kernel.

    size : tuple of int
        The dimensions of the kernel, given as (x_dim, y_dim, z_dim).

    coeffs : list of float, optional
        Coefficients for the Gaussian expression.
        The default is [1, 0, 1, 0, 1, 0], corresponding to the Gaussian form:
        (1 * X^2 + 0 * X * Y + 1 * Y^2 + 0 * Y * Z + 1 * Z^2 + 0 * Z * X).

    Raises
    ------
    ValueError
        If the dimensions in `size` are not positive integers.

    Notes
    -----
    The kernel is generated over a grid that spans twice the size of
    each dimension, and the resulting array is normalized.
    """
    _separable = True
    _is_bool = False

    def __init__(self, stddev, size, coeffs=None):
        if not coeffs:
            coeffs = [1, 0, 1, 0, 1, 0]
        x_dim, y_dim, z_dim = size
        x = np.linspace(-x_dim, x_dim, int(x_dim) + 1)
        y = np.linspace(-y_dim, y_dim, int(y_dim) + 1)
        z = np.linspace(-z_dim, z_dim, int(z_dim) + 1)
        X, Y, Z = np.meshgrid(x, y, z)
        array = np.exp(-(coeffs[0] * X ** 2 +
                         coeffs[1] * X * Y +
                         coeffs[2] * Y ** 2 +
                         coeffs[3] * Y * Z +
                         coeffs[4] * Z ** 2 +
                         coeffs[5] * Z * X) / (2 * stddev ** 2)
                       )
        self._default_size = _round_up_to_odd_integer(stddev)
        super().__init__(array)
        self.normalize()
        self._truncation = np.abs(1. - self._array.sum())


class Interpolator:
    """
    A class to perform data interpolation using convolution with a specified
     kernel.

    Attributes
    ----------
    interp_time : float or None
        Time taken for the last interpolation operation. Defaults to None.

    window : ndarray or None
        Window function to be applied to the interpolated data. Defaults to None.

    interpolated : ndarray or None
        The result of the interpolation operation. Defaults to None.

    data : NXdata or None
        The dataset to be interpolated. Defaults to None.

    kernel : ndarray or None
        The kernel used for convolution during interpolation. Defaults to None.

    tapered : ndarray or None
        The interpolated data after applying the window function. Defaults to None.
    """

    def __init__(self):
        """
        Initialize an Interpolator object.

        Sets up an instance of the Interpolator class with the
        following attributes initialized to None:

        - interp_time
        - window
        - interpolated
        - data
        - kernel
        - tapered

        """
        self.interp_time = None
        self.window = None
        self.interpolated = None
        self.data = None
        self.kernel = None
        self.tapered = None

    def set_data(self, data):
        """
        Set the dataset to be interpolated.

        Parameters
        ----------
        data : NXdata
            The dataset containing the data to be interpolated.
        """
        self.data = data

    def set_kernel(self, kernel):
        """
        Set the kernel to be used for interpolation.

        Parameters
        ----------
        kernel : ndarray
            The kernel to be used for convolution during interpolation.
        """
        self.kernel = kernel

    def interpolate(self):
        """
        Perform interpolation on the dataset using the specified kernel.

        The interpolation is done by convolving the data with the kernel
         using the `convolve_fft` function. Updates the `interpolated`
          attribute with the result.

        Prints the time taken for the interpolation process.

        Returns
        -------
        None
        """
        start = time.time()

        if self.interp_time:
            print(f"Last interpolation took {self.interp_time / 60:.2f} minutes.")

        print("Running interpolation...")
        result = np.real(
            convolve_fft(self.data[self.data.signal].nxdata,
                         self.kernel, allow_huge=True, return_fft=False))
        print("Interpolation finished.")

        end = time.time()
        interp_time = end - start

        print(f'Interpolation took {interp_time / 60:.2f} minutes.')

        result[result < 0] = 0
        self.interpolated = array_to_nxdata(result, self.data)

    def set_tukey_window(self, tukey_alphas=(1.0, 1.0, 1.0)):
        """
        Set a Tukey window function for data tapering.

        Parameters
        ----------
        tukey_alphas : tuple of floats, optional
            The alpha parameters for the Tukey window in each
             dimension (H, K, L). Default is (1.0, 1.0, 1.0).

        Notes
        -----
        The window function is generated based on the size of the dataset in each dimension.
        """
        data = self.data
        tukey_H = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[0]]), alpha=tukey_alphas[0])[:, None, None],
            (1, len(data[data.axes[1]]), len(data[data.axes[2]]))
        )
        tukey_K = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[1]]), alpha=tukey_alphas[1])[None, :, None],
            (len(data[data.axes[0]]), 1, len(data[data.axes[2]]))
        )
        window = tukey_H * tukey_K

        del tukey_H, tukey_K
        gc.collect()

        tukey_L = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[2]]), alpha=tukey_alphas[2])[None, None, :],
            (len(data[data.axes[0]]), len(data[data.axes[1]]), 1))
        window = window * tukey_L

        self.window = window

    def set_hexagonal_tukey_window(self, tukey_alphas=(1.0, 1.0, 1.0, 1.0)):
        """
        Set a hexagonal Tukey window function for data tapering.

        Parameters
        ----------
        tukey_alphas : tuple of floats, optional
            The alpha parameters for the Tukey window in each dimension and
            for the hexagonal truncation (H, HK, K, L).
            Default is (1.0, 1.0, 1.0, 1.0).

        Notes
        -----
        The hexagonal Tukey window is applied to the dataset in a manner that
        preserves hexagonal symmetry.

        """
        data = self.data
        H_ = data[data.axes[0]]
        K_ = data[data.axes[1]]
        L_ = data[data.axes[2]]

        tukey_H = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[0]]), alpha=tukey_alphas[0])[:, None, None],
            (1, len(data[data.axes[1]]), len(data[data.axes[2]]))
        )
        tukey_K = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[1]]), alpha=tukey_alphas[1])[None, :, None],
            (len(data[data.axes[0]]), 1, len(data[data.axes[2]]))
        )
        window = tukey_H * tukey_K

        del tukey_H, tukey_K
        gc.collect()

        truncation = int((len(H_) - int(len(H_) * np.sqrt(2) / 2)) / 2)

        tukey_HK = scipy.ndimage.rotate(
            np.tile(
                np.concatenate(
                    (np.zeros(truncation)[:, None, None],
                     scipy.signal.windows.tukey(len(H_) - 2 * truncation,
                                        alpha=tukey_alphas[2])[:, None, None],
                     np.zeros(truncation)[:, None, None])),
                (1, len(K_), len(L_))
            ),
            angle=45, reshape=False, mode='nearest',
        )[0:len(H_), 0:len(K_), :]
        tukey_HK = np.nan_to_num(tukey_HK)
        window = window * tukey_HK

        del tukey_HK
        gc.collect()

        tukey_L = np.tile(
            scipy.signal.windows.tukey(len(data[data.axes[2]]), alpha=tukey_alphas[3])[None, None, :],
            (len(data[data.axes[0]]), len(data[data.axes[1]]), 1)
        )
        window = window * tukey_L

        del tukey_L
        gc.collect()

        self.window = window

    def set_window(self, window):
        """
        Set a custom window function for data tapering.

        Parameters
        ----------
        window : ndarray
            A custom window function to be applied to the interpolated data.
        """
        self.window = window

    def apply_window(self):
        """
        Apply the window function to the interpolated data.

        The window function, if set, is applied to the `interpolated` data
         to produce the `tapered` result.

        Returns
        -------
        None
        """
        self.tapered = self.interpolated * self.window


def fourier_transform_nxdata(data):
    """
    Perform a 3D Fourier Transform on the given NXdata object.

    This function applies an inverse Fourier Transform to the input data
    using the `pyfftw` library to optimize performance. The result is a
    transformed array with spatial frequency components calculated along
    each axis.

    Parameters
    ----------
    data : NXdata
        An NXdata object containing the data to be transformed. It should
        include the `signal` field for the data and `axes` fields
        specifying the coordinate axes.

    Returns
    -------
    NXdata
        A new NXdata object containing the Fourier Transformed data. The
        result includes:

        - `dPDF`: The transformed data array.
        - `x`, `y`, `z`: Arrays representing the frequency components along each axis.

    Notes
    -----
    - The FFT is performed in two stages: first along the last dimension of
    the input array and then along the first two dimensions.
    - The function uses `pyfftw` for efficient computation of the Fourier
    Transform.
    - The output frequency components are computed based on the step sizes
    of the original data axes.

    """
    start = time.time()
    print("Starting FFT.")

    padded = data[data.signal].nxdata

    fft_array = np.zeros(padded.shape)
    print("FFT on axes 1,2")
    for k in range(0, padded.shape[2]):
        fft_array[:, :, k] = np.real(
            np.fft.fftshift(
                pyfftw.interfaces.numpy_fft.ifftn(np.fft.fftshift(padded[:, :, k]),
                                                  planner_effort='FFTW_MEASURE'))
        )
        print(f'k={k}                  ', end='\r')

    print("FFT on axis 3")
    for i in range(0, padded.shape[0]):
        for j in range(0, padded.shape[1]):
            f_slice = fft_array[i, j, :]
            print(f'i={i}                  ', end='\r')
            fft_array[i, j, :] = np.real(
                np.fft.fftshift(
                    pyfftw.interfaces.numpy_fft.ifftn(np.fft.fftshift(f_slice),
                                                      planner_effort='FFTW_MEASURE')
                )
            )

    end = time.time()
    print("FFT complete.")
    print('FFT took ' + str(end - start) + ' seconds.')

    H_step = data[data.axes[0]].nxdata[1] - data[data.axes[0]].nxdata[0]
    K_step = data[data.axes[1]].nxdata[1] - data[data.axes[1]].nxdata[0]
    L_step = data[data.axes[2]].nxdata[1] - data[data.axes[2]].nxdata[0]

    fft = NXdata(NXfield(fft_array, name='dPDF'),
                 (NXfield(np.linspace(-0.5 / H_step, 0.5 / H_step, padded.shape[0]), name='x'),
                  NXfield(np.linspace(-0.5 / K_step, 0.5 / K_step, padded.shape[1]), name='y'),
                  NXfield(np.linspace(-0.5 / L_step, 0.5 / L_step, padded.shape[2]), name='z'),
                  )
                 )
    return fft


class DeltaPDF:
    """
        A class for processing and analyzing 3D diffraction data using various\
        operations, including masking, interpolation, padding, and Fourier
        transformation.

        Attributes
        ----------
        fft : NXdata or None
            The Fourier transformed data.
        data : NXdata or None
            The input diffraction data.
        lattice_params : tuple or None
            Lattice parameters (a, b, c, al, be, ga).
        reciprocal_lattice_params : tuple or None
            Reciprocal lattice parameters (a*, b*, c*, al*, be*, ga*).
        puncher : Puncher
            An instance of the Puncher class for generating masks and punching
            the data.
        interpolator : Interpolator
            An instance of the Interpolator class for interpolating and applying
            windows to the data.
        padder : Padder
            An instance of the Padder class for padding the data.
        mask : ndarray or None
            The mask used for data processing.
        kernel : Kernel or None
            The kernel used for interpolation.
        window : ndarray or None
            The window applied to the interpolated data.
        padded : ndarray or None
            The padded data.
        tapered : ndarray or None
            The data after applying the window.
        interpolated : NXdata or None
            The interpolated data.
        punched : NXdata or None
            The punched data.

        """

    def __init__(self):
        """
        Initialize a DeltaPDF object with default attributes.
        """
        self.reciprocal_lattice_params = None
        self.fft = None
        self.data = None
        self.lattice_params = None
        self.puncher = Puncher()
        self.interpolator = Interpolator()
        self.padder = Padder()
        self.mask = None
        self.kernel = None
        self.window = None
        self.padded = None
        self.tapered = None
        self.interpolated = None
        self.punched = None

    def set_data(self, data):
        """
        Set the input diffraction data and update the Puncher and Interpolator
         with the data.

        Parameters
        ----------
        data : NXdata
            The diffraction data to be processed.
        """
        self.data = data
        self.puncher.set_data(data)
        self.interpolator.set_data(data)
        self.tapered = data
        self.padded = data
        self.interpolated = data
        self.punched = data

    def set_lattice_params(self, lattice_params):
        """
        Sets the lattice parameters and calculates the reciprocal lattice
         parameters.

        Parameters
        ----------
        lattice_params : tuple of float
            The lattice parameters (a, b, c, alpha, beta, gamma) in real space.
        """
        self.lattice_params = lattice_params
        self.puncher.set_lattice_params(lattice_params)
        self.reciprocal_lattice_params = self.puncher.reciprocal_lattice_params

    def add_mask(self, maskaddition):
        """
         Add regions to the current mask using a logical OR operation.

         Parameters
         ----------
         maskaddition : ndarray
             The mask to be added.
         """
        self.puncher.add_mask(maskaddition)
        self.mask = self.puncher.mask

    def subtract_mask(self, masksubtraction):
        """
        Remove regions from the current mask using a logical AND NOT operation.

        Parameters
        ----------
        masksubtraction : ndarray
            The mask to be subtracted.
        """
        self.puncher.subtract_mask(masksubtraction)
        self.mask = self.puncher.mask

    def generate_bragg_mask(self, punch_radius, coeffs=None, thresh=None):
        """
        Generate a mask for Bragg peaks.

        Parameters
        ----------
        punch_radius : float
            Radius for the Bragg peak mask.
        coeffs : list, optional
            Coefficients for the expression of the sphere to be removed
             around each Bragg position, corresponding to coefficients
              for H, HK, K, KL, L, and LH terms. Default is [1, 0, 1, 0, 1, 0].
        thresh : float, optional
            Intensity threshold for applying the mask.

        Returns
        -------
        mask : ndarray
            Boolean mask identifying the Bragg peaks.
        """
        return self.puncher.generate_bragg_mask(punch_radius, coeffs, thresh)

    def generate_intensity_mask(self, thresh, radius, verbose=True):
        """
        Generate a mask based on intensity thresholds.

        Parameters
        ----------
        thresh : float
            Intensity threshold for creating the mask.
        radius : int
            Radius around high-intensity points to include in the mask.
        verbose : bool, optional
            Whether to print progress information.

        Returns
        -------
        mask : ndarray
            Boolean mask highlighting regions with high intensity.
        """
        return self.puncher.generate_intensity_mask(thresh, radius, verbose)

    def generate_mask_at_coord(self, coordinate, punch_radius, coeffs=None, thresh=None):
        """
        Generate a mask centered at a specific coordinate.

        Parameters
        ----------
        coordinate : tuple of float
            Center coordinate (H, K, L) for the mask.
        punch_radius : float
            Radius for the mask.
        coeffs : list, optional
            Coefficients for the expression of the sphere to be removed around
             each Bragg position, corresponding to coefficients for
             H, HK, K, KL, L, and LH terms. Default is [1, 0, 1, 0, 1, 0].
        thresh : float, optional
            Intensity threshold for applying the mask.

        Returns
        -------
        mask : ndarray
            Boolean mask for the specified coordinate.
        """
        return self.puncher.generate_mask_at_coord(coordinate, punch_radius, coeffs, thresh)

    def punch(self):
        """
        Apply the mask to the dataset, setting masked regions to NaN.

        This method creates a new dataset where the masked regions are set to
         NaN, effectively "punching" those regions.
        """
        self.puncher.punch()
        self.punched = self.puncher.punched
        self.interpolator.set_data(self.punched)

    def set_kernel(self, kernel):
        """
        Set the kernel to be used for interpolation.

        Parameters
        ----------
        kernel : ndarray
            The kernel to be used for convolution during interpolation.
        """
        self.interpolator.set_kernel(kernel)
        self.kernel = kernel

    def interpolate(self):
        """
        Perform interpolation on the dataset using the specified kernel.

        The interpolation is done by convolving the data with the kernel using
         the `convolve_fft` function. Updates the `interpolated` attribute with
          the result.

        Prints the time taken for the interpolation process.

        Returns
        -------
        None
        """
        self.interpolator.interpolate()
        self.interpolated = self.interpolator.interpolated

    def set_tukey_window(self, tukey_alphas=(1.0, 1.0, 1.0)):
        """
        Set a Tukey window function for data tapering.

        Parameters
        ----------
        tukey_alphas : tuple of floats, optional
            The alpha parameters for the Tukey window in each dimension
             (H, K, L). Default is (1.0, 1.0, 1.0).

        Notes
        -----
        The window function is generated based on the size of the dataset
        in each dimension.
        """
        self.interpolator.set_tukey_window(tukey_alphas)
        self.window = self.interpolator.window

    def set_hexagonal_tukey_window(self, tukey_alphas=(1.0, 1.0, 1.0, 1.0)):
        """
        Set a hexagonal Tukey window function for data tapering.

        Parameters
        ----------
        tukey_alphas : tuple of floats, optional
            The alpha parameters for the Tukey window in each dimension and
             for the hexagonal truncation (H, HK, K, L). Default is (1.0, 1.0, 1.0, 1.0).

        Notes
        -----
        The hexagonal Tukey window is applied to the dataset in a manner that
         preserves hexagonal symmetry.
        """
        self.interpolator.set_hexagonal_tukey_window(tukey_alphas)

    def set_window(self, window):
        """
        Set a custom window function for data tapering.

        Parameters
        ----------
        window : ndarray
            A custom window function to be applied to the interpolated data.
        """
        self.interpolator.set_window(window)

    def apply_window(self):
        """
        Apply the window function to the interpolated data.

        The window function, if set, is applied to the `interpolated` data to
         produce the `tapered` result.

        Returns
        -------
        None
        """
        self.interpolator.apply_window()
        self.tapered = self.interpolator.tapered
        self.padder.set_data(self.tapered)

    def pad(self, padding):
        """
        Symmetrically pads the data with zero values.

        Parameters
        ----------
        padding : tuple
            The number of zero-value pixels to add along each edge of the array.

        Returns
        -------
        NXdata
            The padded data with symmetric zero padding.
        """
        self.padded = self.padder.pad(padding)

    def perform_fft(self):
        """
        Perform a 3D Fourier Transform on the padded data.

        This function applies an inverse Fourier Transform to the padded data
        using the `pyfftw` library to optimize performance. The result is a
        transformed array with spatial frequency components calculated along
        each axis.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        - The FFT is performed in two stages: first along the last dimension of
        the input array and then along the first two dimensions.
        - The function uses `pyfftw` for efficient computation of the Fourier
        Transform.
        - The output frequency components are computed based on the step sizes
        of the original data axes.


        """
        self.fft = fourier_transform_nxdata(self.padded)



import matplotlib.pyplot as plt
from nxs_analysis_tools import TempDependence

sample = TempDependence()

sample.load_datasets(folder='../docs/source/examples/example_data/sample_name')

print(sample.datasets)

print(sample.datasets['15'])

sample.datasets['15'][:,:,0.0].plot()

sample.cut_data(center=(0,0,0), window=(0.1,1,0.1))

sample.plot_linecuts()


"""
This module provides classes and functions for analyzing scattering datasets collected at CHESS
(ID4B) with temperature dependence. It includes functions for loading data, cutting data, and
plotting linecuts.
"""
import os
import re

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from IPython.display import display, Markdown
from nxs_analysis_tools import load_data, Scissors
from nxs_analysis_tools.fitting import LinecutModel
from nxs_analysis_tools.datareduction import load_transform, reciprocal_lattice_params


class TempDependence:
    """
    A class for analyzing temperature-dependent scattering datasets collected at CHESS (ID4B).

    The `TempDependence` class facilitates the loading, processing, and analysis of scattering
    data across different temperatures. It includes methods for handling datasets, setting
    lattice parameters, performing linecuts, modeling the data, and visualizing the results.

    Attributes
    ----------
    sample_directory : str
        Path to the directory containing the datasets.
    xlabel : str
        Label for the x-axis of plots, determined by the axis of the linecuts.
    datasets : dict
        Dictionary storing datasets keyed by temperature.
    temperatures : list of str
        List of temperatures for which data is available.
    scissors : dict
        Dictionary of Scissors objects, one for each temperature, used for data manipulation and
        linecut operations.
    linecuts : dict
        Dictionary storing the linecut data for each temperature.
    linecutmodels : dict
        Dictionary of LinecutModel objects, one for each temperature, used for fitting the linecuts.
    a, b, c, al, be, ga : float or None
        Lattice parameters (a, b, c, alpha, beta, gamma) of the crystal.
    a_star, b_star, c_star, al_star, be_star, ga_star : float or None
        Reciprocal lattice parameters (a*, b*, c*, alpha*, beta*, gamma*).

    Methods
    -------
    set_temperatures(temperatures):
        Set the list of temperatures for the datasets.
    set_sample_directory(path):
        Set the directory path where the datasets are located.
    initialize():
        Initialize Scissors and LinecutModel objects for each temperature.
    set_data(temperature, data):
        Set the dataset for a specific temperature.
    load_transforms(temperatures_list=None):
        Load transform datasets (from nxrefine) based on temperature.
    load_datasets(file_ending='hkli.nxs', temperatures_list=None):
        Load datasets (CHESS format) from the specified folder.
    get_sample_directory():
        Get the folder path where the datasets are located.
    clear_datasets():
        Clear the datasets stored in the TempDependence instance.
    set_Lattice_params(lattice_params):
        Set lattice parameters and calculate reciprocal lattice parameters.
    set_window(window, verbose=False):
        Set the extents of the integration window for each temperature.
    set_center(center):
        Set the central coordinate for the linecut for each temperature.
    cut_data(center=None, window=None, axis=None, verbose=False):
        Perform data cutting for each temperature dataset.
    plot_linecuts(vertical_offset=0, **kwargs):
        Plot the linecuts obtained from data cutting.
    plot_linecuts_heatmap(ax=None, **kwargs):
        Plot a heatmap of the linecuts obtained from data cutting.
    highlight_integration_window(temperature=None, **kwargs):
        Display the integration window plot for a specific temperature.
    plot_integration_window(temperature=None, **kwargs):
        Plot the integration window cross-sections for a specific temperature.
    set_model_components(model_components):
        Set the model components for all line cut models.
    set_param_hint(*args, **kwargs):
        Set parameter hints for all line cut models.
    make_params():
        Create parameters for all line cut models.
    guess():
        Make initial parameter guesses for all line cut models.
    print_initial_params():
        Print the initial parameter values for all line cut models.
    plot_initial_guess():
        Plot the initial guess for all line cut models.
    fit(verbose=False):
        Fit the line cut models for each temperature.
    plot_fit(mdheadings=False, **kwargs):
        Plot the fit results for each temperature.
    print_fit_report():
        Print the fit report for each temperature.
    """

    def __init__(self):
        """
        Initialize the TempDependence class with default values.
        """

        self.sample_directory = ''
        self.xlabel = ''
        self.datasets = {}
        self.temperatures = []
        self.scissors = {}
        self.linecuts = {}
        self.linecutmodels = {}
        self.a, self.b, self.c, self.al, self.be, self.ga, \
            self.a_star, self.b_star, self.c_star, self.al_star, self.be_star, self.ga_star \
            = [None] * 12

    def set_temperatures(self, temperatures):
        """
        Set the list of temperatures for the datasets.

        Parameters
        ----------
        temperatures : list
            List of temperatures to set.
        """
        self.temperatures = temperatures

    def set_sample_directory(self, path):
        """
        Set the directory path where the datasets are located.

        Parameters
        ----------
        path : str
            Path to the sample directory.
        """
        self.sample_directory = os.path.normpath(path)

    def initialize(self):
        """
        Initialize Scissors and LinecutModel objects for each temperature.
        """
        for temperature in self.temperatures:
            self.scissors[temperature] = Scissors()
            self.scissors[temperature] = LinecutModel()

    def set_data(self, temperature, data):
        """
        Set the dataset for a specific temperature.

        Parameters
        ----------
        temperature : str
            Temperature for which to set the data.
        data : object
            The dataset to be set.
        """
        self.datasets[temperature] = data

    def load_transforms(self, temperatures_list=None):
        """
        Load transform datasets (from nxrefine) based on temperature.

        Parameters
        ----------
        temperatures_list : list of int or None, optional
            List of temperatures to load. If None, all available temperatures are loaded.
        """
        # Convert all temperatures to strings
        if temperatures_list:
            temperatures_list = [str(t) for t in temperatures_list]

        # Identify files to load
        items_to_load = []
        # Search for nxrefine .nxs files
        for item in os.listdir(self.sample_directory):
            pattern = r'_(\d+)\.nxs'
            match = re.search(pattern, item)
            if match:
                print(f'Found {item}')
                # Identify temperature
                temperature = match.group(1)
                # print(f'Temperature = {temperature}')
                if (temperatures_list is None) or (temperature in temperatures_list):
                    # Prepare file to be loaded
                    self.temperatures.append(temperature)
                    items_to_load.append(item)
                    # print(f'Preparing to load {temperature} K data: {item}')

        # Convert all temperatures to int temporarily to sort temperatures list before loading
        self.temperatures = [int(t) for t in self.temperatures]
        self.temperatures.sort()
        self.temperatures = [str(t) for t in self.temperatures]

        for i, item in enumerate(items_to_load):
            path = os.path.join(self.sample_directory, item)

            # Save dataset
            self.datasets[self.temperatures[i]] = load_transform(path)

            # Initialize scissors object
            self.scissors[self.temperatures[i]] = Scissors()
            self.scissors[self.temperatures[i]].set_data(self.datasets[self.temperatures[i]])

            # Initialize linecutmodel object
            self.linecutmodels[self.temperatures[i]] = LinecutModel()

    def load_datasets(self, file_ending='hkli.nxs', temperatures_list=None):
        """
        Load datasets (CHESS format) from the specified folder.

        Parameters
        ----------
        file_ending : str, optional
            The file extension of the datasets to be loaded. The default is 'hkli.nxs'.
        temperatures_list : list of int or None, optional
            The list of specific temperatures to load. If None, all available temperatures are
            loaded. The default is None.
        """
        temperature_folders = []  # Empty list to store temperature folder names
        for item in os.listdir(self.sample_directory):
            try:
                temperature_folders.append(int(item))  # If folder name can be int, add it
            except ValueError:
                pass  # Otherwise don't add it
        temperature_folders.sort()  # Sort from low to high T
        temperature_folders = [str(i) for i in temperature_folders]  # Convert to strings

        print('Found temperature folders:')
        [print('[' + str(i) + '] ' + folder) for i, folder in enumerate(temperature_folders)]

        self.temperatures = temperature_folders

        if temperatures_list is not None:
            self.temperatures = [str(t) for t in temperatures_list]

        # Load .nxs files
        for T in self.temperatures:
            for file in os.listdir(os.path.join(self.sample_directory, T)):
                if file.endswith(file_ending):
                    filepath = os.path.join(self.sample_directory, T, file)
                    print('-----------------------------------------------')
                    print('Loading ' + T + ' K indexed .nxs files...')
                    print('Found ' + filepath)

                    # Load dataset at each temperature
                    self.datasets[T] = load_data(filepath)

                    # Initialize scissors object at each temperature
                    self.scissors[T] = Scissors()
                    self.scissors[T].set_data(self.datasets[T])

                    # Initialize linecutmodel object at each temperature
                    self.linecutmodels[T] = LinecutModel()

    def get_sample_directory(self):
        """
        Get the folder path where the datasets are located.

        Returns
        -------
        str
            The folder path.
        """
        return self.sample_directory

    def clear_datasets(self):
        """
        Clear the datasets stored in the TempDependence instance.
        """
        self.datasets = {}

    def set_lattice_params(self, lattice_params):
        """
        Set lattice parameters and calculate reciprocal lattice parameters.

        Parameters
        ----------
        lattice_params : tuple
            Tuple containing lattice parameters (a, b, c, al, be, ga).
        """
        self.a, self.b, self.c, self.al, self.be, self.ga = lattice_params
        self.a_star, self.b_star, self.c_star, \
            self.al_star, self.be_star, self.ga_star = reciprocal_lattice_params(lattice_params)

    def set_window(self, window, verbose=False):
        """
        Set the extents of the integration window for each temperature.

        Parameters
        ----------
        window : tuple
            Extents of the window for integration along each axis.
        verbose : bool, optional
            Enables printout of linecut axis and integrated axes. Default is False.
        """
        for T in self.temperatures:
            if verbose:
                print("----------------------------------")
                print("T = " + T + " K")
            self.scissors[T].set_window(window, verbose)

    def set_center(self, center):
        """
        Set the central coordinate for the linecut for each temperature.

        Parameters
        ----------
        center : tuple
            Central coordinate around which to perform the linecut.
        """
        for T in self.temperatures:
            self.scissors[T].set_center(center)

    def cut_data(self, center=None, window=None, axis=None, verbose=False):
        """
        Perform data cutting for each temperature dataset.

        Parameters
        ----------
        center : tuple, optional
            The center point for cutting the data.
            Defaults to the first temperature's center if None.
        window : tuple, optional
            The window size for cutting the data.
            Defaults to the first temperature's window if None.
        axis : int or None, optional
            The axis along which to perform the cutting.
            Defaults to the longest axis in `window` if None.
        verbose : bool, optional
            Enables printout of linecut progress. Default is False.

        Returns
        -------
        dict
            A dictionary of linecuts obtained from the cutting operation.
        """

        center = center if center is not None else self.scissors[self.temperatures[0]].center
        window = window if window is not None else self.scissors[self.temperatures[0]].window
        axis = axis if axis is not None else self.scissors[self.temperatures[0]].axis

        for T in self.temperatures:
            if verbose:
                print("-------------------------------")
                print("Cutting T = " + T + " K data...")
            self.scissors[T].cut_data(center, window, axis, verbose)
            self.linecuts[T] = self.scissors[T].linecut
            self.linecutmodels[T].set_data(self.linecuts[T])

        xlabel_components = [self.linecuts[self.temperatures[0]].axes
                             if i == self.scissors[self.temperatures[0]].axis
                             else str(c) for i, c in
                             enumerate(self.scissors[self.temperatures[0]].center)]
        self.xlabel = ' '.join(xlabel_components)

        return self.linecuts

    def plot_linecuts(self, vertical_offset=0, **kwargs):
        """
        Plot the linecuts obtained from data cutting.

        Parameters
        ----------
        vertical_offset : float, optional
            The vertical offset between linecuts on the plot. The default is 0.
        **kwargs
            Additional keyword arguments to be passed to the plot function.
        """
        fig, ax = plt.subplots()

        # Get the Viridis colormap
        cmap = mpl.colormaps.get_cmap('viridis')

        for i, linecut in enumerate(self.linecuts.values()):
            x_data = linecut[linecut.axes].nxdata
            y_data = linecut[linecut.signal].nxdata + i * vertical_offset
            ax.plot(x_data, y_data, color=cmap(i / len(self.linecuts)), label=self.temperatures[i],
                    **kwargs)

        ax.set(xlabel=self.xlabel,
               ylabel=self.linecuts[self.temperatures[0]].signal)

        # Get the current legend handles and labels
        handles, labels = plt.gca().get_legend_handles_labels()

        # Reverse the order of handles and labels
        handles = handles[::-1]
        labels = labels[::-1]

        # Create a new legend with reversed order
        plt.legend(handles, labels)

        return fig, ax

    def plot_linecuts_heatmap(self, ax=None, **kwargs):
        """
        Plot the linecuts obtained from data cutting.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            The axes on which to plot the heatmap. If None, a new figure and axes
            are created. The default is None.
        **kwargs
            Additional keyword arguments to be passed to the `pcolormesh` function.

        Returns
        -------
        QuadMesh
            The plotted heatmap object.
        """

        # Retrieve linecut data for the first temperature and extract x-axis data
        cut = self.linecuts[self.temperatures[0]]
        x = cut[cut.axes].nxdata

        # Convert the list of temperatures to a NumPy array for the y-axis
        y = np.array([int(t) for t in self.temperatures])

        # Collect counts from each temperature and ensure they are numpy arrays
        v = [self.linecuts[T].counts.nxdata for T in self.temperatures]

        # Convert list of arrays to a 2D array for the heatmap
        v_2d = np.array(v)

        # Create the grid for the heatmap
        X, Y = np.meshgrid(x, y)

        # Plot using pcolormesh
        if ax is None:
            _, ax = plt.subplots()
        p = ax.pcolormesh(X, Y, v_2d, **kwargs)
        plt.colorbar(p, label='counts')
        ax.set(xlabel=cut.axes, ylabel=r'$T$ (K)')

        return p

    def highlight_integration_window(self, temperature=None, **kwargs):
        """
        Displays the integration window plot for a specific temperature,
        or for the first temperature if none is provided.

        Parameters
        ----------
        temperature : str, optional
            The temperature at which to display the integration window plot. If provided, the plot
            will be generated using the dataset corresponding to the specified temperature. If not
            provided, the integration window plots will be generated for the first temperature.
        **kwargs : keyword arguments, optional
            Additional keyword arguments to customize the plot.
        """

        if temperature is not None:
            p = self.scissors[
                self.temperatures[0]].highlight_integration_window(
                data=self.datasets[temperature], **kwargs
            )
        else:
            p = self.scissors[self.temperatures[0]].highlight_integration_window(
                data=self.datasets[self.temperatures[0]], **kwargs
            )

        return p

    def plot_integration_window(self, temperature=None, **kwargs):
        """
        Plots the three principal cross-sections of the integration volume on
        a single figure for a specific temperature, or for the first temperature
        if none is provided.

        Parameters
        ----------
        temperature : str, optional
            The temperature at which to plot the integration volume. If provided,
            the plot will be generated using the dataset corresponding to the
            specified temperature. If not provided, the integration window plots
            will be generated for the first temperature.

        **kwargs : keyword arguments, optional
            Additional keyword arguments to customize the plot.
        """

        if temperature is not None:
            p = self.scissors[self.temperatures[0]].plot_integration_window(**kwargs)
        else:
            p = self.scissors[self.temperatures[0]].plot_integration_window(**kwargs)

        return p

    def set_model_components(self, model_components):
        """
        Set the model components for all line cut models.

        This method sets the same model components for all line cut models in the
        analysis. It iterates over each line cut model and calls their respective
        `set_model_components` method with the provided `model_components`.

        Parameters
        ----------
        model_components : Model or iterable of Model
            The model components to set for all line cut models.

        """
        [linecutmodel.set_model_components(model_components) for
         linecutmodel in self.linecutmodels.values()]

    def set_param_hint(self, *args, **kwargs):
        """
        Set parameter hints for all line cut models.

        This method sets the parameter hints for all line cut models in the analysis.
        It iterates over each line cut model and calls their respective `set_param_hint` method
        with the provided arguments and keyword arguments.

        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.

        """
        [linecutmodel.set_param_hint(*args, **kwargs)
         for linecutmodel in self.linecutmodels.values()]

    def make_params(self):
        """
        Make parameters for all line cut models.

        This method creates the parameters for all line cut models in the analysis.
        It iterates over each line cut model and calls their respective `make_params` method.
        """
        [linecutmodel.make_params() for linecutmodel in self.linecutmodels.values()]

    def guess(self):
        """
        Make initial parameter guesses for all line cut models.

        This method generates initial parameter guesses for all line cut models in the analysis.
        It iterates over each line cut model and calls their respective `guess` method.

        """
        [linecutmodel.guess() for linecutmodel in self.linecutmodels.values()]

    def print_initial_params(self):
        """
        Print the initial parameter values for all line cut models.

        This method prints the initial parameter values for all line cut models
        in the analysis. It iterates over each line cut model and calls their
        respective `print_initial_params` method.

        """
        [linecutmodel.print_initial_params() for linecutmodel in self.linecutmodels.values()]

    def plot_initial_guess(self):
        """
        Plot the initial guess for all line cut models.

        This method plots the initial guess for all line cut models in the analysis.
        It iterates over each line cut model and calls their respective `plot_initial_guess` method.

        """
        for T, linecutmodel in self.linecutmodels.items():
            _, ax = plt.subplots()
            ax.set(title=T + ' K')
            linecutmodel.plot_initial_guess()

    def fit(self, verbose=False):
        """
        Fit the line cut models.

        This method fits the line cut models for each temperature in the analysis.
        It iterates over each line cut model, performs the fit, and prints the fitting progress.

        Parameters
        ----------
       verbose : bool, optional
            Enables printout of fitting progress. Default False.

        """
        for T, linecutmodel in self.linecutmodels.items():
            if verbose:
                print(f"Fitting {T} K  data...")
            linecutmodel.fit()
            if verbose:
                print("Done.")
        print("Fits completed.")

    def plot_fit(self, mdheadings=False, **kwargs):
        """
        Plot the fit results.

        This method plots the fit results for each temperature in the analysis.
        It iterates over each line cut model, calls their respective `plot_fit` method,
        and sets the xlabel, ylabel, and title for the plot.

        """
        for T, linecutmodel in self.linecutmodels.items():
            # Create a markdown heading for the plot
            if mdheadings:
                display(Markdown(f"### {T} K Fit Results"))
            # Plot fit
            linecutmodel.plot_fit(xlabel=self.xlabel,
                                  ylabel=self.datasets[self.temperatures[0]].signal,
                                  title=f"{T} K",
                                  **kwargs)

    def print_fit_report(self):
        """
        Plot the fit results.

        This method plots the fit results for each temperature in the analysis.
        It iterates over each line cut model, calls their respective `plot_fit` method,
        and sets the xlabel, ylabel, and title for the plot.

        """
        for T, linecutmodel in self.linecutmodels.items():
            print(f"[[[{T} K Fit Report]]]")
            linecutmodel.print_fit_report()



from nxs_analysis_tools.datareduction import load_data
from nxs_analysis_tools.pairdistribution import *
import matplotlib.pyplot as plt

data = load_data('../docs/source/examples/example_data/pairdistribution_data/test_hkli.nxs')
# s = Symmetrizer3D(data)

# s.plane1symmetrizer.set_parameters(theta_min=0, theta_max=90, mirror=True)
# s.plane1symmetrizer.test(data[:,:,len(data.L)//2])
#
# s.plane2symmetrizer.set_parameters(theta_min=45, theta_max=90, mirror=True)
# s.plane2symmetrizer.test(data[:,len(data.K)//2,:])
#
# s.plane3symmetrizer.set_parameters(theta_min=0, theta_max=90, mirror=False)
# s.plane3symmetrizer.test(data[len(data.H)//2,:,:])
#
# s.symmetrize()

p = Puncher()
p.set_data(data)
p.set_lattice_params((1,1,1,90,90,90))
bm = p.generate_bragg_mask(punch_radius=0.25)
p.add_mask(bm)
p.punch()
plot_slice(p.punched[:,:,0.0])
plt.show()

m = p.generate_mask_at_coord(coordinate=(0.33, 0.33, 0.0), punch_radius=0.25)
p.add_mask(m)
p.punch()
plot_slice(p.punched[:,:,0.0])
plt.show()

m = p.generate_mask_at_coord(coordinate=(-0.1, -0.1, 0.0), punch_radius=0.2)
p.subtract_mask(m)
p.punch()
plot_slice(p.punched[:,:,0.0])
plt.show()

# SUCCESS


import sys
sys.path.append('../src/nxs_analysis_tools/')
from datareduction import load_data, Scissors, rotate_data, plot_slice

data = load_data('../docs/source/examples/example_data/sample_name/15/example_hkli.nxs')
# scissors = Scissors(data, center=(0,0,0), window=(0.1,2,0.3))
# scissors.cut_data()
# print(scissors.integration_window)
# scissors.plot_integration_window()
# scissors.linecut.plot()
# scissors.highlight_integration_window()

import matplotlib.pyplot as plt

plot_slice(data[:, :, 0.0])
plt.show()
rotated_data = rotate_data(data=data, lattice_angle=90, rotation_angle=45, rotation_axis=2, printout=True)
plot_slice(rotated_data[:, :, 0.0])
plt.show()



import sys
sys.path.append('../src/nxs_analysis_tools/')
from datareduction import load_data, Scissors
from chess import *
from fitting import *

from lmfit.models import GaussianModel, LinearModel, LorentzianModel

sample = TempDependence()
sample.load_datasets(folder=r'K:\pokharel-3470-a\PrCd3P3\BRO7')
sample.cut_data(center=(0.5,-3,0), window=(0.1,0.3,0.1))

sample.set_model_components([GaussianModel(prefix='peak'), LorentzianModel(prefix='diffuse'), LinearModel(prefix='background')])
sample.set_param_hint('peakcenter', min=-0.1, max=0.1)
sample.make_params()
sample.guess()
sample.print_initial_params()
sample.plot_initial_guess()
sample.fit()
sample.plot_fit()
sample.print_fit_report()


'''
Reduce and transform nexus format (.nxs) scattering data.
'''

import numpy as np
from _meta import __author__, __copyright__, __license__, __version__
from .datareduction import load_data, load_transform, plot_slice,\
    reciprocal_lattice_params, Scissors, rotate_data
from .chess import TempDependence

# What to import when running "from nxs_analysis_tools import *"
__all__ = ['load_data', 'load_transform', 'plot_slice', 'Scissors', 'TempDependence',
           'reciprocal_lattice_params', 'rotate_data']



'''nxs-analysis-tools package metadata.'''

# keep consistent with pyproject.toml
__project__ = 'nxs-analysis-tools'
__author__ = 'Steven J. Gomez Alvarado'
__email__ = 'stevenjgomez@ucsb.edu'
__copyright__ = f"2023, {__author__}"
__license__ = 'MIT'
__version__ = '0.0.39'
__repo_url__ = 'https://github.com/stevenjgomez/nxs_analysis_tools'
